'''
Known Peripheral UUIDs, obtained by querying using the Bluepy module:
=====================================================================
Anti DOS Characteristic <00020005-574f-4f20-5370-6865726f2121>
Battery Level Characteristic <Battery Level>
Peripheral Preferred Connection Parameters Characteristic <Peripheral Preferred Connection Parameters>
API V2 Characteristic <00010002-574f-4f20-5370-6865726f2121>
DFU Control Characteristic <00020002-574f-4f20-5370-6865726f2121>
Name Characteristic <Device Name>
Appearance Characteristic <Appearance>
DFU Info Characteristic <00020004-574f-4f20-5370-6865726f2121>
Service Changed Characteristic <Service Changed>
Unknown1 Characteristic <00020003-574f-4f20-5370-6865726f2121>
Unknown2 Characteristic <00010003-574f-4f20-5370-6865726f2121>

The rest of the values saved in the dictionaries below, were borrowed from
@igbopie's javacript library, which is available at https://github.com/igbopie/spherov2.js

'''

deviceID = {"apiProcessor": 0x10,                   # 16
            "systemInfo": 0x11,                     # 17
            "powerInfo": 0x13,                      # 19
            "driving": 0x16,                        # 22
            "animatronics": 0x17,                   # 23
            "sensor": 0x18,                         # 24
            "something": 0x19,                      # 25
            "userIO": 0x1a,                         # 26
            "somethingAPI": 0x1f}                   # 31

SystemInfoCommands = {"mainApplicationVersion": 0x00,   # 00
                      "bootloaderVersion": 0x01,    # 01
                      "something": 0x06,            # 06
                      "something2": 0x13,           # 19
                      "something6": 0x12,           # 18    
                      "something7": 0x28}           # 40

sendPacketConstants = {"StartOfPacket": 0x8d,       # 141
                       "EndOfPacket": 0xd8}         # 216

userIOCommandIDs = {"allLEDs": 0x0e}                # 14

flags= {"isResponse": 0x01,                         # 0x01
        "requestsResponse": 0x02,                   # 0x02
        "requestsOnlyErrorResponse": 0x04,          # 0x04
        "resetsInactivityTimeout": 0x08}            # 0x08

powerCommandIDs={"deepSleep": 0x00,                 # 0
                "sleep": 0x01,                      # 01
                "batteryVoltage": 0x03,             # 03
                "wake": 0x0D,                       # 13
                "something": 0x05,                  # 05 
                "something2": 0x10,                 # 16         
                "something3": 0x04,                 # 04
                "something4": 0x1E}                 # 30

drivingCommands={"rawMotor": 0x01,                  # 1
                 "resetHeading": 0x06,              # 6    
                 "driveAsSphero": 0x04,             # 4
                 "driveAsRc": 0x02,                 # 2
                 "driveWithHeading": 0x07,          # 7
                 "stabilization": 0x0C}             # 12

sensorCommands={'sensorMask': 0x00,                 # 00
                'sensorResponse': 0x02,             # 02
                'configureCollision': 0x11,         # 17
                'collisionDetectedAsync': 0x12,     # 18
                'resetLocator': 0x13,               # 19
                'enableCollisionAsync': 0x14,       # 20
                'sensor1': 0x0F,                    # 15
                'sensor2': 0x17,                    # 23
                'configureSensorStream': 0x0C}      # 12



class linear_trend_tracker():
    """ linear trend tracker with adaptive forgetting factor
    """   

    def __init__(self,halflife=70,int_err_halflife=5, K_int_err=10, a0=1, b0=0):
        self.alpha=np.exp(np.log(.5)/halflife) if halflife else .99
        self.warmup_weight= (1-self.alpha**20)/(1-self.alpha); # >20 points for warmup
        self.N=0
        if int_err_halflife is None and halflife:
            int_err_halflife = max(halflife/100,1) 
        self.int_err_alpha = np.exp(np.log(.5)/int_err_halflife) if int_err_halflife else .999
        self.K_int_err = K_int_err
        self.a0 = a0
        self.b0 = b0

    def reset(self, keep_err=False, keep_model=False):
        self.N = 0
        self.X0 = None
        self.Y0 = None
        self.sX = 0
        self.sY = 0
        self.sXX = 0
        self.sYX = 0
        self.sYY = 0
        if not keep_model:
            self.a=self.a0
            self.b=self.b0
        if not keep_err:
            self.int_err = 0
            self.abs_err=0
            self.int_err_N = 0

    def fit(self,X,Y,keep_err=False):
        self.reset(keep_err)
        if hasattr(X,'__iter__'):
            self.N = len(X)
            self.X0=X[0,...] 
            self.Y0=Y[0,...] 
            if len(X)>1 :
                self.transform(X[1:,...],Y[1:,...])
        else:
            self.N = 1
            self.X0 = X
            self.Y0 = Y
        self.a = self.a0
        self.b = self.Y0 - self.a * self.X0

    def transform(self,X,Y):
        if self.N==0:
            self.fit(X,Y)
            return Y

        #if np.all(X==self.Xlast) or np.all(Y==self.Ylast):
        #    return self.getY(X)
        # get our prediction for this point and use to track our prediction error
        Yest = self.getY(X)
        err  = np.mean(np.abs(Y-Yest))
  
        ## center x/y
        # N.B. be sure to do all the analysis in floating point to avoid overflow
        cX  = np.array(X - self.X0, dtype=float)
        cY  = np.array(Y - self.Y0, dtype=float)
        N = len(X) if hasattr(X,'__iter__') else 1
        # update the 1st and 2nd order summary statistics 
        wght    = self.alpha**N
        # adaptive learning rate as function of integerated error
        if self.N > self.warmup_weight:
            ptwght = max(1, abs(self.int_err/self.int_err_N)*self.K_int_err) #1/max(1,err) if self.N > self.warmup_weight else 1
        else:
            ptwght = 1
        # adaptive learning rate as a function of the direction of the error...
        #if err < 0:
        #    ptwght = ptwght*.1
        self.N  = wght*self.N   + ptwght*N
        self.sY = wght*self.sY  + ptwght*np.sum(cY,axis=0)
        self.sX = wght*self.sX  + ptwght*np.sum(cX,axis=0)
        self.sYY= wght*self.sYY + ptwght*np.sum(cY*cY,axis=0)
        self.sYX= wght*self.sYX + ptwght*np.sum(cY*cX,axis=0)
        self.sXX= wght*self.sXX + ptwght*np.sum(cX*cX,axis=0)

        # update the slope when warmed up
        if self.N > self.warmup_weight:
            Yvar  = self.sYY - self.sY * self.sY / self.N
            Xvar  = self.sXX - self.sX * self.sX / self.N
            YXvar = self.sYX - self.sY * self.sX / self.N
            self.a = (YXvar / Xvar + Yvar/YXvar )/2        
        # update the bias given the estimated slope b = mu_y - a * mu_x
        # being sure to include the shift to the origin!
        self.b = (self.Y0 + self.sY / self.N) - self.a * (self.X0 + self.sX / self.N)

        # check for steps in the inputs
        # get our prediction for this point and use to track our prediction error
        Yest = self.getY(X)
        err  = np.mean(Y-Yest)

        # track the prediction error, with long and short half-life
        self.abs_err = self.abs_err*wght + abs(err)
        # track with step window
        int_err_wght = self.int_err_alpha**N
        self.int_err_N = self.int_err_N * int_err_wght + N
        self.int_err = self.int_err*int_err_wght + err

        # only return the estimate if we've warmed up the tracker
        if self.N < self.warmup_weight:
            return Y

        # detect change in statistics by significant difference in error statistics
        # between long and short halflife
        #if (self.step_err / self.step_N) > (self.err / self.N) * self.step_threshold:
        #    print("step-detected")
        #    #self.fit(X,Y,keep_err=True)
        #    #Yest = Y

        return Yest

    def getX(self,y):
        return ( y  - self.b ) / self.a 

    def getY(self,x):
        return self.a * x + self.b

    @staticmethod
    def testcase():
        X = np.arange(1000) + np.random.randn(1)*1e6
        a = 1000/50 
        b = 1000*np.random.randn(1)
        Ytrue= a*X+b
        Y    = Ytrue+ np.random.standard_normal(Ytrue.shape)*10

        import glob
        import os
        files = glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)),'../../logs/mindaffectBCI*.txt')) # * means all if need specific format then *.csv
        savefile = max(files, key=os.path.getctime)
        #savefile = "C:\\Users\\Developer\\Downloads\\mark\\mindaffectBCI_brainflow_200911_1339.txt" 
        #savefile = "C:/Users/Developer/Downloads/khash/mindaffectBCI_brainflow_ipad_200908_1938.txt"
        from mindaffectBCI.decoder.offline.read_mindaffectBCI import read_mindaffectBCI_messages
        from mindaffectBCI.utopiaclient import DataPacket
        print("Loading: {}".format(savefile))
        msgs = read_mindaffectBCI_messages(savefile,regress=None) # load without time-stamp fixing.
        dp = [ m for m in msgs if isinstance(m,DataPacket)]
        nsc = np.array([ (m.samples.shape[0],m.sts,m.timestamp) for m in dp])
        X = np.cumsum(nsc[:,0])
        Y = nsc[:,1]
        Ytrue = nsc[:,2]

        ltt = linear_trend_tracker(1000)
        ltt.fit(X[0],Y[0]) # check scalar inputs
        step = 1
        idxs = list(range(1,X.shape[0],step))
        ab = np.zeros((len(idxs),2))
        print("{}) a={} b={}".format('true',a,b))
        dts = np.zeros((Y.shape[0],))
        dts[0] = ltt.getY(X[0])
        for i,j in enumerate(idxs):
            dts[j:j+step] = ltt.transform(X[j:j+step],Y[j:j+step])
            ab[i,:] = (ltt.a,ltt.b)
            yest = ltt.getY(X[j])
            err =  yest - Y[j]
            if abs(err)> 1000:
                print("{}) argh! yest={} ytrue={} err={}".format(i,yest,Ytrue[j],err))
            if i < 100:
                print("{:4d}) a={:5f} b={:5f}\ty_est-y={:2.5f}".format(j,ab[i,0],ab[i,1],
                        Y[j]-yest))

        import matplotlib.pyplot as plt
        ab,res,_,_ = np.linalg.lstsq(np.append(X[:,np.newaxis],np.ones((X.shape[0],1)),1),Y,rcond=-1)
        ots = X*ab[0]+ab[1]        
        idx=range(X.shape[0])
        plt.plot(X[idx],Y[idx]- X[idx]*ab[0]-Y[0],label='server ts')
        plt.plot(X[idx],dts[idx] - X[idx]*ab[0]-Y[0],label='regressed ts (samp vs server)')
        plt.plot(X[idx],ots[idx] - X[idx]*ab[0]-Y[0],label='regressed ts (samp vs server) offline')

        err = Y - X*ab[0] - Y[0]
        cent = np.median(err); scale=np.median(np.abs(err-cent))
        plt.ylim((cent-scale*5,cent+scale*5))

        plt.legend()
        plt.show()

if __name__ == "__main__":
    linear_trend_tracker.testcase('-')



import numpy as np
from mindaffectBCI.decoder.updateSummaryStatistics import updateSummaryStatistics, zero_outliers, crossautocov, updateCyy, updateCxy, autocov, updateCxx, plot_erp, plot_summary_statistics, plot_factoredmodel
from mindaffectBCI.decoder.multipleCCA import multipleCCA
from mindaffectBCI.decoder.scoreOutput import scoreOutput, dedupY0
from mindaffectBCI.decoder.utils import window_axis
from mindaffectBCI.decoder.scoreStimulus import scoreStimulus, scoreStimulusCont, factored2full
from mindaffectBCI.decoder.decodingCurveSupervised import decodingCurveSupervised
from mindaffectBCI.decoder.decodingSupervised import decodingSupervised
from mindaffectBCI.decoder.stim2event import stim2event
from mindaffectBCI.decoder.normalizeOutputScores import normalizeOutputScores, estimate_Fy_noise_variance
from mindaffectBCI.decoder.zscore2Ptgt_softmax import calibrate_softmaxscale

try:
    from sklearn.model_selection import StratifiedKFold
    from sklearn.base import BaseEstimator, ClassifierMixin
    from sklearn.exceptions import NotFittedError
except:
#  if True:
    # placeholder classes for when sklearn isn't available
    class StratifiedKFold:
        def __init__(self, n_splits):
            self.n_splits = n_splits

        def split(self, X, Y):
            shape = X.shape
            self.splits = np.linspace(0, shape[0], self.n_splits+1, dtype=int, endpoint=True).tolist()
            self._index = 0
            return self

        def __iter__(self):
            return self

        def __next__(self):
            if self._index < len(self.splits)-1:
                train_idx = list(range(self.splits[self._index]))+list(range(self.splits[self._index+1], self.splits[-1]))
                val_idx = list(range(self.splits[self._index], self.splits[self._index+1]))
                self._index = self._index + 1
            else:
                raise StopIteration
            return (train_idx, val_idx)


    class BaseEstimator:
        pass


    class ClassifierMixin:
        pass

    class NotFittedError(Exception):
        pass
    
class BaseSequence2Sequence(BaseEstimator, ClassifierMixin):
    '''Base class for sequence-to-sequence learning.  Provides, prediction and scoring functions, but not the fitting method'''
    def __init__(self, evtlabs=('re','fe'), tau=18, offset=0, outputscore='ip'):
        self.evtlabs = evtlabs if evtlabs is not None else ('re','fe')
        self.tau = tau
        self.offset = offset
        self.outputscore = outputscore
        self.verb = 0
        self.priorweight = 120
        if self.offset>0 or self.offset<-tau:
            raise NotImplementedError("Offsets of more than a negative window are not supported yet!")
        
    def stim2event(self, Y, prevY=None):
        '''transform Stimulus-encoded to brain-encoded, if needed'''
        # convert from stimulus coding to brain response coding
        if self.evtlabs is not None:
            Y = stim2event(Y, self.evtlabs, axis=-2, oM=prevY) # (tr, samp, Y, e)
        else:
            if Y.ndim == 3:
                Y = Y[:,:,:,np.newaxis] # (tr,samp,Y,e)

        return Y

    def fit(self, X, Y):
        '''fit model mapping 2 multi-dim time series: X = (tr, samp, d), Y = (tr, samp, e)'''
        raise NotImplemented

    def is_fitted(self):
        return hasattr(self,"W_")

    def predict(self, X, Y, dedup0=None, prevY=None):
        '''make predictions on multi-dim time series: X = (tr, samp, d), Y = (tr, samp, e)
        
        N.B. this implementation assumes linear coefficients in W_ (nM,nfilt,d) and R_ (nM,nfilt,nE,tau)'''
        if not self.is_fitted():
            # only if we've been fitted!
            raise NotFittedError
        # convert from stimulus coding to brain response coding
        Y = self.stim2event(Y, prevY)
        # valid performance
        Fe = self.transform(X) # (nM, nTrl, nSamp, nE)
        Fy = scoreOutput(Fe, Y, outputscore=self.outputscore, dedup0=dedup0, R=self.R_, offset=self.offset) #(nM, nTrl, nSamp, nY)
        # BODGE: strip un-needed model dimension
        if Fy.shape[0] == 1 and Fy.ndim > 3:
            Fy = Fy.reshape(Fy.shape[1:])
        return Fy

    def transform(self, X):
        ''' estimate the stimulus properties from the raw data '''
        return scoreStimulus(X, self.W_, self.R_, self.b_, offset=self.offset)

    def decode_proba(self, Fy, minDecisLen=0, marginalizemodels=True):
        # build optional arguments if set
        kwargs=dict()
        if hasattr(self,'sigma0_') and self.sigma0_ is not None:
            # include the score prior
            #print('priorsigma=({},{})'.format(self.sigma0_,self.priorweight))
            kwargs['priorsigma']=(self.sigma0_,self.priorweight)
        if hasattr(self,'softmaxscale_') and self.softmaxscale_ is not None:
            kwargs['softmaxscale']=self.softmaxscale_

        Yest, Perr, Ptgt, _, _ = decodingSupervised(Fy, minDecisLen=minDecisLen, marginalizemodels=marginalizemodels, **kwargs)
        return Ptgt #(nTrl, nEp, nY)

    
    def predict_proba(self, X, Y, marginalizemodels=False, minDecisLen=-1, dedup0=None, prevY=None):
        Fy = self.predict(X, Y, dedup0=dedup0, prevY=prevY)
        return self.decode_proba(Fy,marginalizemodels=marginalizemodels, minDecisLen=minDecisLen)

    def score(self, X, Y):
        '''score this model on this data, N.B. for model_selection higher is *better*'''
        # score, removing duplicates for true target (objId==0) so can compute score
        Fy = self.predict(X, Y, dedup0=True) # (nM, nTrl, nSamp, e)

        return BaseSequence2Sequence.audc_score(Fy)

    @staticmethod
    def audc_score(Fy):
        '''compute area under decoding curve score from Fy, *assuming* Fy[:,:,0] is the *true* classes score'''
        sFy = np.cumsum(Fy, axis=-2) # (nM, nTrl, nSamp, nY)
        Yi  = np.argmax(sFy, axis=-1) # output for every model*trial*sample
        score = np.sum((Yi == 0).ravel())/Yi.size # total amount time was right, higher=better
        return score

    def cv_fit(self, X, Y, cv=5, fit_params={}, verbose=0, return_estimator=True, calibrate_softmax=True, dedup0=True):
        ''' cross validated fit to the data.  N.B. write our own as sklearn doesn't work for getting the estimator values for structured output.'''
        # TODO [] : make more computationally efficient by pre-computing the updateSummaryStatistics etc.
        # TODO [] : conform to sklearn cross_validate signature
        # TODO [] : move into a wrapper class
        if cv == True:  cv = 5
        if isinstance(cv, int):
            if X.shape[0] > 1:
                cv = StratifiedKFold(n_splits=min(cv, X.shape[0])).split(np.zeros(X.shape[0]), np.zeros(Y.shape[0]))
            else: # single trial, train/test on all...
                cv = [(slice(1), slice(1))] # N.B. use slice to preserve dims..

        if return_estimator:
            Fy = np.zeros(Y.shape) if Y.ndim<=3 else np.zeros(Y.shape[:-1]) # (nTrl, nEp, nY)
        else:
            Fy = None
        scores = []
        if verbose > 0:
            print("CV:", end='')
        for i, (train_idx, valid_idx) in enumerate(cv):
            if verbose > 0:
                print(".", end='', flush=True)
            #print("trn={} val={}".format(train_idx,valid_idx))
            
            self.fit(X[train_idx, ...], Y[train_idx, ...], **fit_params)

            # predict, forcing removal of copies of  tgt=0 so can score
            Fyi = self.predict(X[valid_idx, ...], Y[valid_idx, ...], dedup0=dedup0)
            if return_estimator:
                Fy[valid_idx, ...] = Fyi
            
            scores.append(self.audc_score(Fyi))

        # final retrain with all the data
        self.fit(X, Y)

        self.sigma0_ = None
        self.softmaxscale_ = None
        if return_estimator:
            # estimate prior for Fy using the cv'd Fy
            #self.sigma0_ = np.sum(Fy.ravel()**2) / Fy.size
            #print('Fy={}'.format(Fy.shape))
            # N.B. need to match the filter used in the decoder..
            self.sigma0_, _ = estimate_Fy_noise_variance(Fy, priorsigma=None)  # per-trial
            #print('Sigma0{} = {}'.format(self.sigma0_.shape,self.sigma0_))
            self.sigma0_ = np.median(self.sigma0_.ravel())  # ave
            print('Sigma0 = {}'.format(self.sigma0_))

            if calibrate_softmax:
                # calibrate the softmax scale to give more valid probabilities
                nFy,_,_,_,_ = normalizeOutputScores(Fy, minDecisLen=-10, priorsigma=(self.sigma0_,self.priorweight))
                self.softmaxscale_ = calibrate_softmaxscale(nFy)

        return {'estimator': Fy, 'test_score': scores}
    
    def plot_model(self, **kwargs):
        if not self.R_ is None:
            print("Plot Factored Model")
            if hasattr(self, 'A_'):
                plot_factoredmodel(self.A_, self.R_, evtlabs=self.evtlabs, spatial_filter_type='Pattern', **kwargs)
            else:
                plot_factoredmodel(self.W_, self.R_, evtlabs=self.evtlabs, **kwargs)
        else:
            plot_erp(self.W_, evtlabs=self.evtlabs, **kwargs)

    
class MultiCCA(BaseSequence2Sequence):
    ''' Sequence 2 Sequence learning using CCA as a bi-directional forward/backward learning method '''
    def __init__(self, evtlabs=('re','fe'), tau=18, offset=0, rank=1, reg=1e-8, rcond=1e-4, badEpThresh=6, symetric=False, center=True, CCA=True, **kwargs):
        super().__init__(evtlabs=evtlabs, tau=tau,  offset=offset, **kwargs)
        self.rank = rank
        self.reg  = reg
        self.rcond= rcond
        self.badEpThresh = badEpThresh
        self.symetric = symetric
        self.center = center
        self.CCA = CCA

    def fit(self, X, Y, stimTimes=None):
        '''fit 2 multi-dim time series: X = (tr, samp, d), Y = (tr, samp, Y)'''
        # map to event sequence
        Y = self.stim2event(Y) # (tr,samp,nY,e)
        # extract the true target to fit to, using horible slicing trick
        Y_true = Y[..., 0:1, :] #  (tr,samp,1,e)
        # get summary statistics
        #print("X={} |X|={}".format(X.shape,np.sum(X**2,axis=(0,1))))
        Cxx, Cxy, Cyy = updateSummaryStatistics(X, Y_true, stimTimes, tau=self.tau, offset=self.offset, badEpThresh=self.badEpThresh, center=self.center)
        #print("diag(Cxx)={}".format(np.diag(Cxx)))
        # do the CCA fit
        J, W, R = multipleCCA(Cxx, Cxy, Cyy, reg=self.reg, rank=self.rank, rcond=self.rcond, CCA=self.CCA, symetric=self.symetric)
        # maintain type compatiability
        W = W.astype(X.dtype)
        R = R.astype(X.dtype)
        self.W_ = W #(nM,rank,d)
        self.R_ = R #(nM,rank,e,tau)
        # store A_ for plotting later
        self.A_ = np.einsum("de,Mkd->Mke",Cxx,W)
        
        if self.center: # use bias terms to correct for data centering
            muX = np.mean(X.reshape((-1,X.shape[-1])),0)
            self.b_ = -np.einsum("Mkd,d,Mket->Me",W,muX,R) # (nM,d)
            self.b_ = self.b_[0,:] # strip model dim..
        else:
            self.b_ = None

        return self
    
class FwdLinearRegression(BaseSequence2Sequence):
    ''' Sequence 2 Sequence learning using forward linear regression  X = A*Y '''
    def __init__(self, evtlabs=('re','fe'), tau=18, offset=0, reg=None, rcond=1e-6, badEpThresh=6, center=True, **kwargs):
        super().__init__(evtlabs=evtlabs, tau=tau, offset=offset, **kwargs)
        self.reg = reg
        self.rcond = rcond
        self.badEpThresh = badEpThresh
        self.center = center
        if offset != 0:
            print("WARNING: not tested with offset!! YMMV")

    def fit(self, X, Y):
        '''fit 2 multi-dim time series: X = (tr, samp, d), Y = (tr, samp, e)'''
        # map to event sequence
        Y = self.stim2event(Y) # (tr,samp,nY,e)
        # extract the true target to fit to
        Y_true = Y[..., 0, :] # (tr,samp, e)
        # get summary statistics
        # first clean up the data..
        X, Y_true = zero_outliers(X, Y_true, self.badEpThresh)
        if self.center:
            # TODO[]: make work in-place w/o modify X
            X = X.copy()
            muX = np.mean(X.reshape((-1,X.shape[-1])),0)
            X = X - muX
        # TODO: should be negative tau to represent causality of Y->X, so earlier Y cause X
        Cyteyte = crossautocov(Y_true, Y_true, tau=self.tau) # (tau,e,tau,e) cross-cov shifted Y with itself
        # N.B. should be shifted Y with X, but we want Y *backwards* in time, thus
        #      as a *BODGE* we instead window X, as forward for X == backwards for Y
        #Cytexd  = crossautocov(Y_true, X, tau=[-self.tau, 1]) # (tau,e,1,d) shifted Y with X
        Cytexd  = crossautocov(Y_true, X, tau=[1, self.tau], offset=self.offset) # (1,e,tau,d) shifted X with Y
        Cytexd  = np.moveaxis(Cytexd, (0, 1, 2, 3), (2, 1, 0, 3)) # (tau,e,1,d)
        # fit:
        # Make 2d
        # (tau,e,tau,e) -> ((tau*e),(tau*e))
        Cyy2d = Cyteyte.reshape((Cyteyte.shape[0]*Cyteyte.shape[1],
                                 Cyteyte.shape[2]*Cyteyte.shape[3]))
        # (tau,e,1,d) -> ((tau*e),d)
        Cyx2d = Cytexd.reshape( (Cytexd.shape[0]*Cytexd.shape[1],
                                 Cytexd.shape[3]))
        # add regularization
        if self.reg:
            # **IN-PLACE modify Cyy2d..**
            diag=Cyy2d.diagonal()
            np.fill_diagonal(Cyy2d, diag+self.reg*np.max(diag))
        
        # fit by solving the matrix equation: Cyy*A=Cyx -> A=Cyy**-1Cyx
        # Q: solve or lstsq?
        #R = np.linalg.solve(Cyy2d, Cyx2d) # ((tau*e),d)
        R,_,_,_ = np.linalg.lstsq(Cyy2d, Cyx2d, rcond=self.rcond) # ((tau*e),d)
        # convert back to 2d (tau,e,d)
        R = R.reshape((Cyteyte.shape[0], Cyteyte.shape[1], R.shape[1]))
        # BODGE: now moveaxis to make consistent with the prediction functions
        self.W_ = np.moveaxis(R, (0, 1, 2), (1, 0, 2)) #(tau,e,d)->(e,tau,d)
        self.A_ = self.W_
        self.R_ = None
        if self.center:
            self.b_ = -np.einsum("etd,d->e", self.W_, muX) #(e,)
        else:
            self.b_ = None

        return self

class BwdLinearRegression(BaseSequence2Sequence):
    ''' Sequence 2 Sequence learning using backward linear regression  W*X = Y '''
    def __init__(self, evtlabs=('re','fe'), tau=18, offset=0, reg=None, rcond=1e-5, badEpThresh=6, center=True, **kwargs):
        super().__init__(evtlabs=evtlabs, tau=tau, offset=offset, **kwargs)
        self.reg = reg
        self.rcond = rcond
        self.badEpThresh = badEpThresh
        self.center = center
        if offset != 0:
            print("WARNING: not tested with offset!! YMMV")

    def fit(self, X, Y):
        '''fit 2 multi-dim time series: X = (tr, samp, d), Y = (tr, samp, e)'''
        # map to event sequence
        Y = self.stim2event(Y) # (tr,samp,nY,e)
        # extract the true target to fit to
        Y = Y[..., 0, :] # (tr,samp, e)
        # first clean up the data..
        X, Y = zero_outliers(X, Y, self.badEpThresh)
        if self.center:
            # TODO[]: make work in-place w/o modify X
            X = X.copy()
            muX = np.mean(X.reshape((-1,X.shape[-1])),0)
            X = X - muX
        # get summary statistics
        Cxtdxtd = crossautocov(X, X, tau=self.tau) # (tau,d,tau,d) cross-cov shifted X with itself
        Cxtdye  = crossautocov(X, Y, tau=[self.tau, 1], offset=self.offset) # (tau,d,1,e) shifted X with Y
        # fit:
        # Make 2d
        # (tau,d,tau,d) -> ((tau*d),(tua*d))
        Cxx2d = Cxtdxtd.reshape((Cxtdxtd.shape[0]*Cxtdxtd.shape[1],
                                 Cxtdxtd.shape[2]*Cxtdxtd.shape[3]))
        # (tau,d,1,e) -> ((tau*d),e)
        Cxy2d = Cxtdye.reshape( (Cxtdye.shape[0]*Cxtdye.shape[1],
                                 Cxtdye.shape[3]))
        # add regularization
        if self.reg:
            # **IN-PLACE modify Cxx2d..**
            diag = Cxx2d.diagonal()
            np.fill_diagonal(Cxx2d, diag+self.reg*np.max(diag))
        
        # fit by solving the matrix equation: Cxx*W=Cxy -> W=Cxx**-1Cxy
        W,_,_,_ = np.linalg.lstsq(Cxx2d, Cxy2d, rcond=self.rcond) # ((tau*d*),e)
        #W = np.linalg.solve(Cxx2d, Cxy2d) # ((tau*d*),e) # can have problems with singular inputs!
        # convert back to 2d
        W = W.reshape((Cxtdxtd.shape[0], Cxtdxtd.shape[1], W.shape[1])) # (tau,d,e)
        # BODGE: moveaxis to make consistent for the prediction functions
        W = np.moveaxis(W, (0, 1, 2), (1, 2, 0)) #(tau,d,e)->(e,tau,d)
        self.W_ = W
        self.R_ = None
        if self.center:
            self.b_ = -np.einsum("etd,d->e",self.W_,muX) #(e,)
        else:
            self.b_ = None

        return self

#from sklearn.linear_model import Ridge
class LinearSklearn(BaseSequence2Sequence):
    ''' Wrap a normal sk-learn classifier for sequence to sequence learning '''
    def __init__(self, clsfr, labelizeY=False, ignore_unlabelled=True, badEpThresh=None, **kwargs):
        super().__init__(**kwargs)
        self.clsfr = clsfr
        self.labelizeY = labelizeY
        self.badEpThresh = badEpThresh
        self.ignore_unlabelled = ignore_unlabelled
        
    @staticmethod
    def sklearn_fit(X, Y, clsfr, tau, offset, labelizeY=False, ignore_unlabelled=True, badEpThresh=None, verb=0):
        '''fit 2 multi-dim time series: X = (tr, samp, d), Y = (tr, samp, e) using a given sklearn linear_model method'''
        # slice X to get block of data per sample
        X = window_axis(X, winsz=tau, axis=1) # (tr,samp-tau,tau,d)
        # sub-set Y to  be the same size
        Y = Y[:, offset:X.shape[1]+offset, :] # (tr,samp-tau,e)

        # clean up the data
        if badEpThresh is not None:
            X, Y = zero_outliers(X, Y, badEpThresh)

        # convert to a 2d array: (tr*samp, feat) before giving to sklearn
        X2d = np.reshape(X, (np.prod(X.shape[:2]), np.prod(X.shape[2:]))) #((tr*samp-tau),(tau*d))
        Y2d = np.reshape(Y, (np.prod(Y.shape[:2]), np.prod(Y.shape[2:]))) #((tr*samp-tau),e)
        if labelizeY: # convert Y to a label list
            Ylab = np.argmax(Y2d, axis=-1) # ((tr*samp),1)
            unlabelled = np.all(Y2d == 0, axis=-1)
            if np.any(unlabelled):
                if ignore_unlabelled:
                    if verb > 0: print("Warning: removing unlabelled examples from training")
                    keep = ~unlabelled
                    Ylab= Ylab[keep]
                    Y2d = Y2d[keep, ...]
                    X2d = X2d[keep, ...]
                    if verb > 0: print("X={} Y={} Ylab={}".format(X2d.shape,Y2d.shape,np.unique(Ylab)))
                else:
                    if verb > 0: print("Warning: added an extra 'REST' class")
                    # add extra class for anything which isn't labelled already
                    Ylab[unlabelled] = Y2d.shape[-1]
                
            Y2d = Ylab
        # fit the classifier
        clsfr.fit(X2d, Y2d)

        # log the binary training set performance.
        if verb > 0: print("Training score = {}".format(clsfr.score(X2d,Y2d)))

        return clsfr

    def fit(self, X, Y):
        # map to event sequence
        Y = self.stim2event(Y)
        # extract the true target to fit to
        Y_true = Y[..., 0, :] #  (tr,samp,e)
        # fit the model
        self.clsfr = LinearSklearn.sklearn_fit(X, Y_true, self.clsfr, tau=self.tau, offset=self.offset, labelizeY=self.labelizeY, ignore_unlabelled=self.ignore_unlabelled, badEpThresh=self.badEpThresh)
        # extract the solution coefficient and store in W_
        W = self.clsfr.coef_ # ( n_class, n_feat ) = (e,(tau*d)))
        b = self.clsfr.intercept_ # (e,1)
        #print("W={}".format(W.shape))
        #print("b={}".format(b.shape))
        W = np.reshape(W, (W.shape[0] if W.ndim > 1 else 1, self.tau, X.shape[-1])) # (e,tau,d)
        #print("W={}".format(W.shape))

        # ensure the weights vector is the right size ...
        if self.labelizeY and not W.shape[0] == Y_true.shape[-1]:

            if W.shape[0] > Y.shape[-1]:
                if self.verb > 0: print("More weights than outputs! Y={} W={} removed".format(Y.shape[-1],W.shape[0]))
                W = W[:Y.shape[-1], ...]
                b = b[:Y.shape[-1]]
            elif W.shape[0] == 1 and Y.shape[-1] == 2:
                if self.verb > 0: print("Fewer weights than outputs! Y={} W={} binary".format(Y.shape[-1],W.shape[0]))
                # Y is binary, so e==0 -> neg-class,  e==1 -> pos-class
                # TODO[]: check?
                W = np.concatenate((-W, W), axis=0)
                b = np.concatenate((-b, b), axis=0)
            else:
                raise ValueError("Don't know what to do with W")

        self.W_ = W
        self.R_ = None
        self.b_ = b
        #print("b={}".format(b))

        return self
    
def visualize_Fy_Py(Fy, Py):
    import numpy as np
    import matplotlib.pyplot as plt
    plt.ion()
    for trli in range(Fy.shape[0]):
        print('{})'.format(trli), flush=True)
        plt.clf()
        plt.subplot(211);plt.plot(np.cumsum(Fy[trli, :, :], -2))
        plt.subplot(212);plt.plot(Py[trli, :, :])
        plt.title("{})".format(trli))
        plt.draw()
        plt.pause(1)

def plot_model_weights(model:BaseSequence2Sequence, ch_names=None):
    from updateSummaryStatistics import plot_erp
    from scoreStimulus import factored2full
    plot_erp(factored2full(model.W_, model.R_), ch_names=ch_names, evtlabs=model.evtlabs, offset=model.offset)

def testLeadLag():
    import numpy as np
    from utils import testSignal
    from model_fitting import MultiCCA    
    from decodingCurveSupervised import decodingCurveSupervised
    import matplotlib.pyplot as plt

    irf=(0,0,0,0,0,0,0,0,0,1)
    offset=0; # X->lag-by-10
    X,Y,st,A,B = testSignal(nTrl=1,nSamp=500,d=10,nE=1,nY=30,isi=10,tau=10,offset=offset,irf=irf,noise2signal=0)

    # reference case lagged response test
    evtlabs=None; tau=10; cca_offset=0
    cca = MultiCCA(evtlabs=evtlabs, tau=tau, offset=cca_offset)
    scores=cca.cv_fit(X,Y)
    Fy = scores['estimator']
    (_) = decodingCurveSupervised(Fy)
        
    # leading X
    irf=(1,0,0,0,0,0,0,0,0,0)
    offset=-9; # X leads by 9
    X,Y,st,A,B = testSignal(nTrl=1,nSamp=5000,d=10,nE=1,nY=30,isi=10,tau=10,offset=offset,irf=irf,noise2signal=0)
    plt.figure(0);plt.clf();plt.plot(X[0,:,0],label='X');plt.plot(Y[0,:,0,0],label='Y');plt.title("offset={}, irf={}".format(offset,irf));plt.legend()

    # no-shift in analysis window
    evtlabs=None; tau=10; cca_offset=0
    cca = MultiCCA(evtlabs=evtlabs, tau=tau,offset=cca_offset)
    scores=cca.cv_fit(X,Y)
    Fy = scores['estimator']
    (_) = decodingCurveSupervised(Fy)

    # shifted analysis window
    evtlabs=None; tau=20; cca_offset=-9
    cca = MultiCCA(evtlabs=evtlabs, tau=tau,offset=cca_offset)
    scores=cca.cv_fit(X,Y)
    Fy = scores['estimator']
    (_) = decodingCurveSupervised(Fy)

    plt.figure(1);plt.clf();plot_model_weights(cca)
    
        
def testcase(dataset='toy',loader_args=dict()):
    from model_fitting import MultiCCA, FwdLinearRegression, BwdLinearRegression, LinearSklearn
    from decodingCurveSupervised import decodingCurveSupervised
    from datasets import get_dataset
    
    loadfn, filenames, dataroot = get_dataset(dataset)
    if dataset=='toy':
        loader_args=dict(tau=10,isi=5,noise2signal=3,nTrl=20,nSamp=50)
    X, Y, coords = loadfn(filenames[0], **loader_args)
    fs = coords[1]['fs']
    ch_names = coords[2]['coords']

    Y = Y[..., 0] # (nsamp, nY)
        
    # raw
    tau = int(.3*fs)
    evtlabs = ('re', 'fe', 'rest')#  ('re', 'ntre')#  ('re', 'ntre', 'rest')#  ('1', '0')#
    cca = MultiCCA(tau=tau, evtlabs=evtlabs)
    print('cca = {}'.format(cca))
    cca.fit(X, Y)
    Fy = cca.predict(X, Y, dedup0=True)
    print("score={}".format(cca.score(X,Y)))
    (_) = decodingCurveSupervised(Fy)

    # cca - cv-fit
    print("CV fitted")
    cca = MultiCCA(tau=tau, rank=1, reg=None, evtlabs=evtlabs)
    cv_res = cca.cv_fit(X, Y)
    Fy = cv_res['estimator']
    (_) = decodingCurveSupervised(Fy,priorsigma=(cca.sigma0_,cca.priorweight))
        
    from model_fitting import MultiCCA, FwdLinearRegression, BwdLinearRegression
    from updateSummaryStatistics import updateSummaryStatistics, plot_erp, plot_summary_statistics
    from scoreStimulus import factored2full
    import matplotlib.pyplot as  plt
    from decodingCurveSupervised import decodingCurveSupervised
    tau = int(.3*fs)
    rank = 1
    evtlabs = ('re', 'fe', 'rest') #('re', 'ntre') # ('re', 'fe')  # ('1', '0') #
    # cca
    cca = MultiCCA(tau=tau, rank=rank, evtlabs=evtlabs, reg=None)
    print("{}".format(cca))
    cca.fit(X, Y)
    Fy = cca.predict(X, Y, dedup0=True)
    (_) = decodingCurveSupervised(Fy)

    plot_erp(factored2full(cca.W_, cca.R_), ch_names=ch_names, evtlabs=evtlabs)
    plt.savefig('W_cca.png')
    
    # fwd-model
    print("Forward Model")
    fwd = FwdLinearRegression(tau=tau, evtlabs=evtlabs, reg=None, badEpThresh=4)
    print("{}".format(fwd))
    fwd.fit(X, Y)
    Fy = fwd.predict(X, Y, dedup0=True)
    (_) = decodingCurveSupervised(Fy)
    
    # bwd-model
    print("Backward Model")
    bwd = BwdLinearRegression(tau=tau, evtlabs=evtlabs, reg=None, badEpThresh=4)
    print("{}".format(bwd))
    bwd.fit(X, Y)
    Fy = bwd.predict(X, Y, dedup0=True)
    (_) = decodingCurveSupervised(Fy)

    Py = bwd.predict_proba(X, Y, dedup0=True)
    visualize_Fy_Py(Fy, Py)

    # sklearn wrapper
    from sklearn.linear_model import Ridge, LogisticRegression
    from sklearn.svm import LinearSVR, LinearSVC
    print("sklearn-ridge")
    ridge = LinearSklearn(tau=tau, evtlabs=evtlabs, clsfr=Ridge(alpha=0))
    print("{}".format(ridge))
    ridge.fit(X, Y)
    Fy = ridge.predict(X, Y, dedup0=True)
    (_) = decodingCurveSupervised(Fy)


    print("sklear-lr")
    lr = LinearSklearn(tau=tau, evtlabs=evtlabs, clsfr=LogisticRegression(C=1, multi_class='multinomial', solver='sag'), labelizeY=True)
    print("{}".format(lr))
    lr.fit(X, Y)
    Fy = lr.predict(X, Y, dedup0=True)
    (_) = decodingCurveSupervised(Fy)

    print("sklear-svc")
    svc = LinearSklearn(tau=tau, evtlabs=evtlabs, clsfr=LinearSVC(C=1, multi_class='ovr'), labelizeY=True)
    print("{}".format(svc))
    svc.fit(X, Y)
    Fy = svc.predict(X, Y, dedup0=True)
    (_) = decodingCurveSupervised(Fy)

    
    plot_erp(factored2full(svc.W_, svc.R_), ch_names=ch_names, evtlabs=evtlabs)
    plt.savefig('W_svc.png')

    
    # hyper-parameter optimization with cross-validation
    from sklearn.model_selection import GridSearchCV
    tuned_parameters={'rank':[1, 2, 3, 5], 'tau':[int(dur*fs) for dur in [.2, .3, .5, .7]], 'evtlabs':[['re', 'fe'], ['re', 'ntre'], ['0','1']]}
    cv_cca = GridSearchCV(MultiCCA(), tuned_parameters)
    cv_cca.fit(X, Y)
    print("CVOPT:\n\n{} = {}\n".format(cv_cca.best_estimator_, cv_cca.best_score_))
    means = cv_cca.cv_results_['mean_test_score']
    stds = cv_cca.cv_results_['std_test_score']
    for mean, std, params in zip(means, stds, cv_cca.cv_results_['params']):
        print("{:5.3f} (+/-{:5.3f}) for {}".format(mean, std * 2, params))
    print()

    # use the best setting to fit a normal model, and get it's cv estimated predictions
    cca = MultiCCA()
    cca.set_params(**cv_cca.best_params_) # Note: **dict -> k,v argument array
    cv_res = cca.cv_fit(X, Y)
    Fy = cv_res['estimator']
    (_) = decodingCurveSupervised(Fy)
    
    # Slice data
    from utils import sliceData, sliceY
    stimTimes = st[st < X.shape[1]-tau+1] # limit valid stimTimes
    Xe = sliceData(X, stimTimes, tau=tau) # d x tau x ep x trl
    Ye = sliceY(Y, stimTimes, featdim=False) # y x ep  x trl
    print('cca = {}'.format(cca))
    Fy = cca.predict(X, Y)
    print("Fy={}".format(Fy.shape))
    Py = cca.predict_proba(X, Y)
    print("Py={}".format(Py.shape))
    score = cca.score(X, Y)
    print("score={}".format(score))
    decodingCurveSupervised(Fy)
    
if __name__ ==  "__main__":
    testcase()



import numpy as np

# time-series tests
def window_axis(a, winsz, axis=0, step=1, prependwindowdim=False):
    ''' efficient view-based slicing of equal-sized equally-spaced windows along a selected axis of a numpy nd-array '''
    if axis < 0: # no negative axis indices
        axis = len(a.shape)+axis

    # compute the shape/strides for the  windowed view of a
    if prependwindowdim: # window dim before axis
        shape  = a.shape[:axis] + (winsz, int((a.shape[axis]-winsz)/step)+1) +  a.shape[(axis+1):]
        strides =  a.strides[:axis] + (a.strides[axis], a.strides[axis]*step) + a.strides[(axis+1):]
    else: # window dim after axis
        shape  = a.shape[:axis] + (int((a.shape[axis]-winsz)/step)+1, winsz) +  a.shape[(axis+1):]
        strides =  a.strides[:axis] + (a.strides[axis]*step, a.strides[axis]) + a.strides[(axis+1):]
    #print("a={}".format(a.shape))
    #print("shape={} stride={}".format(shape,strides))
        
    # return the computed view
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)

def equals_subarray(a, pat, axis=-1, match=-1):
    ''' efficiently find matches of a 1-d sub-array along axis within an nd-array ''' 
    if axis < 0: # no negative dims
        axis = a.ndim+axis
    # reshape to match dims of a
    if not isinstance(pat, np.ndarray): pat = np.array(pat) # ensure is numpy
    pshape = np.ones(a.ndim+1, dtype=int); pshape[axis+1] = pat.size
    pat =  np.array(pat.ravel()).reshape(pshape) # [ ... x l x...]
    # window a into pat-len pieces
    aw = window_axis(a, pat.size, axis=axis) # [ ... x t-l x l x ...]
    # do the match
    F  = np.all(np.equal(aw, pat), axis=axis+1) # [... x t-l x ...]
    # pad to make the same shape as input
    padshape = list(a.shape); padshape[axis] = a.shape[axis]-F.shape[axis]
    if match == -1: # match at end of pattern -> pad before
        F  = np.append(np.zeros(padshape), F, axis)
    else: # match at start of pattern -> pad after
        F  = np.append(F, np.zeros(padshape), axis)
    return F


class RingBuffer:
    ''' time efficient linear ring-buffer for storing packed data, e.g. continguous np-arrays '''
    def __init__(self, maxsize, shape, dtype=np.float32):
        self.elementshape = shape
        self.bufshape = (int(maxsize), )+shape
        self.buffer = np.zeros((2*int(maxsize), np.prod(shape)), dtype=dtype) # store as 2d
        # position for the -1 element. N.B. start maxsize so pos-maxsize is always valid
        self.pos = int(maxsize)
        self.n = 0 # count of total number elements added to the buffer
        self.copypos = 0 # position of the last element copied to the 1st half
        self.copysize = 0 # number entries to copy as a block

    def clear(self):
        '''empty the ring-buffer and reset to empty'''
        self.pos=int(self.bufshape[0])
        self.n  =0
        self.copypos=0
        self.copysize=0

    def append(self, x):
        '''add single element to the ring buffer'''
        return self.extend(x[np.newaxis, ...])
    
    def extend(self, x):
        '''add a group of elements to the ring buffer'''
        # TODO[] : incremental copy to the 1st half, to spread the copy cost?
        nx = x.shape[0]
        if self.pos+nx >= self.buffer.shape[0]:
            flippos = self.buffer.shape[0]//2
            # flippos-nx to 1st half
            self.buffer[:(flippos-nx), :] = self.buffer[(self.pos-(flippos-nx)):self.pos, :]
            # move cursor to end 1st half
            self.pos = flippos-nx

        # insert in the buffer
        self.buffer[self.pos:self.pos+nx, :] =  x.reshape((nx, self.buffer.shape[1]))
        # move the cursor
        self.pos = self.pos+nx
        # update the count
        self.n = self.n + nx
        return self
    
    @property
    def shape(self):
        return (min(self.n,self.bufshape[0]),)+self.bufshape[1:]
    
    def unwrap(self):
        '''get a view on the valid portion of the ring buffer'''
        return self.buffer[self.pos-min(self.n,self.bufshape[0]):self.pos, :].reshape(self.shape)

    def __getitem__(self, item):
        return self.unwrap()[item]

    def __iter__(self):
        return iter(self.unwrap())

def extract_ringbuffer_segment(rb, bgn_ts, end_ts=None):
    ''' extract the data between start/end time stamps, from time-stamps contained in the last channel of a nd matrix'''
    # get the data / msgs from the ringbuffers
    X = rb.unwrap() # (nsamp,nch+1)
    X_ts = X[:, -1] # last channel is timestamps
    # TODO: binary-search to make these searches more efficient!
    # search backwards for trial-start time-stamp
    # TODO[X] : use a bracketing test.. (better with wrap-arround)
    bgn_samp = np.flatnonzero(np.logical_and(X_ts[:-1] < bgn_ts, bgn_ts <= X_ts[1:]))
    # get the index of this timestamp, guarding for after last sample
    bgn_samp = bgn_samp[0] if len(bgn_samp) > 0 else len(X_ts)+1
    # and just to be sure the trial-end timestamp
    if  end_ts is not None:
        end_samp = np.flatnonzero(np.logical_and(X_ts[:-1] < end_ts, end_ts <= X_ts[1:]))
        # get index of this timestamp, guarding for after last data sample
        end_samp = end_samp[-1] if len(end_samp) > 0 else len(X_ts)
    else: # until now
        end_samp = len(X_ts)
    # extract the trial data, and make copy (just to be sure)
    X = X[bgn_samp:end_samp+1, :].copy()
    return X

def unwrap(x,range=None):
    ''' unwrap a list of numbers to correct for truncation due to limited bit-resolution, e.g. time-stamps stored in 24bit integers'''
    if range is None: 
        range = 1<< int(np.ceil(np.log2(max(x))))
    wrap_ind = np.diff(x) < -range/2
    unwrap = np.zeros(x.shape)
    unwrap[np.flatnonzero(wrap_ind)+1]=range
    unwrap=np.cumsum(unwrap)
    x = x + unwrap
    return x

def unwrap_test():
    x = np.cumsum(np.random.rand(6000,1))
    xw = x%(1<<10)
    xuw = unwrap(x)
    import matplotlib.pyplot as plt
    plt.plot(x,label='x')
    plt.plot(xw,label='x (wrapped)')
    plt.plot(xuw,label='x (unwrapped')
    plt.legend()

# toy data generation

#@function
def randomSummaryStats(d=10, nE=2, tau=10, nY=1):
    import numpy as np
    # pure random test-case
    Cxx = np.random.standard_normal((d, d))
    Cxy = np.random.standard_normal((nY, nE, tau, d))
    Cyy = np.random.standard_normal((nY, nE, tau, nE, tau))
    return (Cxx, Cxy, Cyy)

def testNoSignal(d=10, nE=2, nY=1, isi=5, tau=None, nSamp=10000, nTrl=1):
    # Simple test-problem -- no real signal
    if tau is None:
        tau = 10*isi
    X = np.random.standard_normal((nTrl, nSamp, d))
    stimTimes_samp = np.arange(0, X.shape[-2] - tau, isi)
    Me = np.random.standard_normal((nTrl, len(stimTimes_samp), nY, nE))>1
    Y = np.zeros((nTrl, X.shape[-2], nY, nE))
    Y[:, stimTimes_samp, :, :] = Me
    return (X, Y, stimTimes_samp)

def testSignal(nTrl=1, d=5, nE=2, nY=30, isi=5, tau=None, offset=0, nSamp=10000, stimthresh=.6, noise2signal=1, irf=None):
    #simple test problem, with overlapping response
    import numpy as np
    if tau is None:
        tau = 10 if irf is None else len(irf)
    nEp = int((nSamp-tau)/isi)
    cb = np.random.standard_normal((nEp, nY, nE)) > stimthresh # codebook = per-epoch stimulus activity
    E  = cb # (nEp, nY, nE) # per-epoch stimulus activity
    # up-sample to sample rate
    stimTimes_samp = np.arange(0, nSamp-tau, isi) # (nEp)
    Y = np.zeros((nSamp, nY, E.shape[-1]))
    Y[stimTimes_samp, :, :] = E[:len(stimTimes_samp), :, :] #per-sample stimulus activity (nSamp, nY, nE) [nE x nY x nSamp]
    Y = np.tile(Y,(nTrl,1,1,1)) # replicate for the trials
    # generate the brain source
    A  = np.random.standard_normal((nE, d)) # spatial-pattern for the source signal
    if irf is None:
        B  = np.zeros((tau),dtype=np.float32)
        B[-3] = 1;         # true response filter (shift by 10 samples)
    else:
        B = np.array(irf,dtype=np.float32)
    Ytrue = Y[..., 0, :] # (nTrl, nSamp, nE)

    if True:
        # convolve with the impulse response - manually using window_axis
        # zero pad before for the sliding window
        Ys = np.zeros(Ytrue.shape[:-2]+(Ytrue.shape[-2]+tau-1,)+Ytrue.shape[-1:])
        Ys[..., tau-1+offset:Ytrue.shape[-2]+tau-1+offset, :] = Ytrue # zero-pad at front + include the offset.
        Yse = window_axis(Ys, winsz=len(B), axis=-2) # (nTr,nSamp,tau,nE)
        YtruecB = np.einsum("Tste,t->Tse", Yse, B[::-1]) # N.B. time-reverse irf (nTr,nSamp,nE)

    else:
        # use the np convolve function, N.B. implicitly time reverses B (like we want)
        YtruecB = np.array([np.convolve(Ytrue[:, ei], B, 'full') for ei in range(Ytrue.shape[-1])]).T #(nSamp+pad, nE) [nE x nSamp]
        YtruecB = YtruecB[:Ytrue.shape[0], :] # trim the padding

    #import matplotlib.pyplot as plt; plt.clf(); plt.plot(Ytrue[:100,0],'b*',label='Y'); plt.plot(YtruecB[:100,0],'g*',label='Y*B'); plt.plot(B,'k',label='B'); plt.legend()

    #print("Ytrue={}".format(Ytrue.shape))
    #print("YtruecB={}".format(YtruecB.shape))
    S  = YtruecB # (nTr, nSamp, nE) true response, i.e. filtered Y 
    N  = np.random.standard_normal(S.shape[:-1]+(d,)) # EEG noise (nTr, nSamp, d)
    X  = np.einsum("tse,ed->tsd", S, A) + noise2signal*N       # simulated data.. true source mapped through spatial pattern (nSamp, d) #[d x nSamp]
    return (X, Y, stimTimes_samp, A, B)

def testtestSignal():
    import matplotlib.pyplot as plt
    plt.clf()
    # shift by 5
    offset=0; irf=(0,0,0,0,0,1,0,0,0,0)
    X,Y,st,W,R = testSignal(nTrl=1,nSamp=500,d=1,nE=1,nY=1,isi=10,tau=10,offset=offset,irf=irf,noise2signal=0)
    plt.subplot(311);plt.plot(X[0,:,0],label='X');plt.plot(Y[0,:,0,0],label='Y');plt.title("offset={}, irf={}".format(offset,irf));plt.legend()
    # back-shift-by-5 -> 0 shift
    offset=-5
    X,Y,st,W,R = testSignal(nTrl=1,nSamp=500,d=1,nE=1,nY=1,isi=10,tau=10,offset=offset,irf=(0,0,0,0,0,1,0,0,0,0),noise2signal=0)
    plt.subplot(312);plt.plot(X[0,:,0],label='X');plt.plot(Y[0,:,0,0],label='Y');plt.title("offset={}, irf={}".format(offset,irf));plt.legend()
    # back-shift-by-10 -> -5 shift
    offset=-9
    X,Y,st,W,R = testSignal(nTrl=1,nSamp=500,d=1,nE=1,nY=1,isi=10,tau=10,offset=offset,irf=(0,0,0,0,0,1,0,0,0,0),noise2signal=0)
    plt.subplot(313);plt.plot(X[0,:,0],label='X');plt.plot(Y[0,:,0,0],label='Y');plt.title("offset={}, irf={}".format(offset,irf));plt.legend()
    

def sliceData(X, stimTimes_samp, tau=10):
    # make a sliced version
    dst = np.diff(stimTimes_samp)
    if np.all(dst == dst[0]) and stimTimes_samp[0] == 0: # fast path equaly spaced stimTimes
        Xe = window_axis(X, winsz=tau, axis=-2, step=int(dst[0]), prependwindowdim=False) # (nTrl, nEp, tau, d) #d x tau x ep x trl
    else: 
        Xe = np.zeros(X.shape[:-2] + (len(stimTimes_samp), tau, X.shape[-1])) # (nTrl, nEp, tau, d) [ d x tau x nEp x nTrl ]
        for ei, si in enumerate(stimTimes_samp):
            idx = range(si, si+tau)
            Xe[:, ei, :, :] = X[:, idx, :] if X.ndim > 2 else X[idx, :]
    return Xe

def sliceY(Y, stimTimes_samp, featdim=True):
    '''
    Y = (nTrl, nSamp, nY, nE) if featdim=True
    OR
    Y=(nTrl, nSamp, nY) if featdim=False #(nE x nY x nSamp x nTrl)
    '''
    # make a sliced version
    si = np.array(stimTimes_samp, dtype=int)
    if featdim:
        return Y[:, si, :, :] if Y.ndim > 3 else Y[si, :, :]
    else:
        return Y[:, si, :] if Y.ndim > 2 else Y[si, :]

    
def block_randomize(true_target, npermute, axis=-3, block_size=None):
    ''' make a block random permutaton of the input array
    Inputs:
       npermute: int - number permutations to make
       true_target: (..., nEp, nY, e): true target value for nTrl trials of length nEp flashes
       axis : int the axis along which to permute true_target'''
    if true_target.ndim < 3:
        raise ValueError("true target info must be at least 3d")
    if not (axis == -3 or axis == true_target.ndim-2):
        raise NotImplementedError("Only implementated for axis=-2 currently")
    
    # estimate the number of blocks to use
    if block_size is None:
        block_size = max(1, true_target.shape[axis]/2/npermute)

    nblk = int(np.ceil(true_target.shape[axis]/block_size))
    blk_lims = np.linspace(0, true_target.shape[axis], nblk, dtype=int)
    # convert to start/end index for each block
    blk_lims = [(blk_lims[i], blk_lims[i+1]) for i in range(len(blk_lims)-1)]
    cb = np.zeros(true_target.shape[:axis+1] + (npermute, true_target.shape[-1]))
    for ti in range(cb.shape[axis+1]):
        for di, dest_blk in enumerate(blk_lims):
            
            yi = np.random.randint(true_target.shape[axis+1])
            si = np.random.randint(len(blk_lims))
            # ensure can't be the same block
            if si == di:
                si = si+1 if si < len(blk_lims)-1 else si-1
            src_blk = blk_lims[si]
            # guard for different lengths for source/dest blocks
            dest_len = dest_blk[1] - dest_blk[0]
            if dest_len > src_blk[1]-src_blk[0]:
                if src_blk[0]+dest_len < true_target.shape[axis]:
                    # enlarge the src
                    src_blk = (src_blk[0], src_blk[0]+dest_len)
                elif src_blk[1]-dest_len > 0:
                    src_blk = (src_blk[1]-dest_len, src_blk[1])
                else:
                    raise ValueError("can't fit source and dest")
            elif dest_len < src_blk[1]-src_blk[0]:
                src_blk = (src_blk[0], src_blk[0]+dest_len)
            
            cb[..., dest_blk[0]:dest_blk[1], ti, :] = true_target[..., src_blk[0]:src_blk[1], yi, :]

    return cb


def upsample_codebook(trlen, cb, ep_idx, stim_dur_samp, offset_samp=(0, 0)):
    ''' upsample a codebook definition to sample rate
    Inputs:
       trlen : (int) length after up-sampling
       cb : (nTr, nEp, ...) the codebook
       ep_idx : (nTr, nEp) the indices of the codebook  entries
       stim_dur_samp: (int) the amount of time the cb entry is held for
       offset_samp : (2,):int the offset for the stimulus in the upsampled trlen data
    Outputs:
       Y : ( nTrl, trlen, ...) the up-sampled codebook '''

    if ep_idx is not None:
        if not np.all(cb.shape[:ep_idx.ndim] == ep_idx.shape):
            raise ValueError("codebook and epoch indices must has same shape")
        trl_idx = ep_idx[:, 0] # start each trial
    else: # make dummy ep_idx with 0 for every trial!
        ep_idx = np.zeros((cb.shape[0],1),dtype=int)
        trl_idx = ep_idx
        
    Y  = np.zeros((cb.shape[0], trlen)+ cb.shape[2:], dtype='float32') # (nTr, nSamp, ...)
    for ti, trl_start_idx in enumerate(trl_idx):
        for ei, epidx in enumerate(ep_idx[ti, :]):
            if ei > 0 and epidx == 0: # zero indicates end of variable length trials
                break
            # start index for this epoch in this *trial*, including the 0-offset 
            ep_start_idx = -int(offset_samp[0])+int(epidx-trl_start_idx)
            Y[ti, ep_start_idx:(ep_start_idx+int(stim_dur_samp)), ...]  = cb[ti, ei, ...]
    return Y


def lab2ind(lab,lab2class=None):
    ''' convert a list of labels (as integers) to a class indicator matrix'''
    if lab2class is None:
        lab2class = [ (l,) for l in set(lab) ] # N.B. list of lists
    if not isinstance(lab,np.ndarray):
        lab=np.array(lab)
    Y = np.zeros(lab.shape+(len(lab2class),),dtype=bool)
    for li,ls in enumerate(lab2class):
        for l in ls:
            Y[lab == l, li]=True
    return (Y,lab2class)


def zero_outliers(X, Y, badEpThresh=4, badEpChThresh=None, verbosity=0):
    '''identify and zero-out bad/outlying data

    Inputs:
      X = (nTrl, nSamp, d)
      Y = (nTrl, nSamp, nY, nE) OR (nTrl, nSamp, nE)
               nE=#event-types  nY=#possible-outputs  nEpoch=#stimulus events to process
    '''
    # remove whole bad epochs first
    if badEpThresh > 0:
        bad_ep, _ = idOutliers(X, badEpThresh, axis=(-2, -1)) # ave over time,ch
        if np.any(bad_ep):
            if verbosity > 0:
                print("{} badEp".format(np.sum(bad_ep.ravel())))
            # copy X,Y so don't modify in place!
            X = X.copy()
            Y = Y.copy()
            X[bad_ep[..., 0, 0], ...] = 0
            #print("Y={}, Ybad={}".format(Y.shape, Y[bad_ep[..., 0, 0], ...].shape))
            # zero out Y also, so don't try to 'fit' the bad zeroed data
            Y[bad_ep[..., 0, 0], ...] = 0
            
            
    # Remove bad individual channels next
    if badEpChThresh is None: badEpChThresh = badEpThresh*2 
    if badEpChThresh > 0:
        bad_epch, _ = idOutliers(X, badEpChThresh, axis=-2) # ave over time
        if np.any(bad_epch):
            if verbosity > 0:
                print("{} badEpCh".format(np.sum(bad_epch.ravel())))
            # make index expression to zero out the bad entries
            badidx = list(np.nonzero(bad_epch)) # convert to linear indices
            badidx[-2] = slice(X.shape[-2]) # broadcast over the accumulated dimensions
            if not np.any(bad_ep): # copy so don't update in place
                X = X.copy()
            X[tuple(badidx)] = 0

    return (X, Y)


def idOutliers(X, thresh=4, axis=-2, verbosity=0):
    ''' identify outliers with excessively high power in the input data
    Inputs:
      X:float the data to identify outliers in
      axis:int (-2)  axis of X to sum to get power
      thresh(float): threshold standard deviation for outlier detection
      verbosity(int): verbosity level
    Returns:
      badEp:bool (X.shape axis==1) indicator for outlying elements
      epPower:float (X.shape axis==1) power used to identify bad
    '''
    #print("X={} ax={}".format(X.shape,axis))
    power = np.sqrt(np.sum(X**2, axis=axis, keepdims=True))  
    #print("power={}".format(power.shape))
    good = np.ones(power.shape, dtype=bool)
    for _ in range(4):
        mu = np.mean(power[good])
        sigma = np.sqrt(np.mean((power[good] - mu) ** 2))
        badThresh = mu + thresh*sigma
        good[power > badThresh] = False
    good = good.reshape(power.shape) # (nTrl, nEp)
    #print("good={}".format(good.shape))
    bad = ~good
    if verbosity > 1:
        print("%d bad" % (np.sum(bad.ravel())))
    return (bad, power)

try:
    from scipy.signal import butter, bessel, sosfilt, sosfilt_zi
except:
#if True:
    # use the pure-python fallbacks
    def sosfilt(sos,X,axis,zi):
        return sosfilt_2d_py(sos,X,axis=axis,zi=zi)
    def sosfilt_zi(sos):
        return sosfilt_zi_py(sos)
    def butter(order,freq,btype,output):
        return butter_py(order,freq,btype,output)

def sosfilt_zi_warmup(zi, X, axis=-1, sos=None):
    '''Use some initial data to "warmup" a second-order-sections filter to reduce startup artifacts.

    Args:
        zi (np.ndarray): the sos filter, state
        X ([type]): the warmup data
        axis (int, optional): The filter axis in X. Defaults to -1.
        sos ([type], optional): the sos filter coefficients. Defaults to None.

    Returns:
        [np.ndarray]: the warmed up filter coefficients
    '''
    
    if axis < 0: # no neg axis
        axis = X.ndim+axis
    # zi => (order,...,2,...)
    zi = np.reshape(zi, (zi.shape[0],) + (1,)*(axis) + (zi.shape[1],) + (1,)*(X.ndim-axis-1))

    # make a programattic index expression to support arbitary axis
    idx = [slice(None)]*X.ndim
    # get the index to start the warmup
    warmupidx = 0 if sos is None else min(sos.size*3,X.shape[axis]-1)

    # center on 1st warmup value
    idx[axis] = slice(warmupidx,warmupidx+1)
    zi = zi * X[tuple(idx)] 

    # run the filter on the rest of the warmup values
    if not sos is None and warmupidx>3: 
        idx[axis] = slice(warmupidx,1,-1)
        _, zi  = sosfilt(sos, X[tuple(idx)], axis=axis, zi=zi)
        
    return zi

def iir_sosfilt_sos(stopband, fs, order=4, ftype='butter', passband=None, verb=0):
    ''' given a set of filter cutoffs return butterworth or bessel sos coefficients '''

    # convert to normalized frequency, Note: not to close to 0/1
    if stopband is None:
        return np.array(())

    if not hasattr(stopband[0],'__iter__'):
        stopband=(stopband,)

    sos=[]
    for sb in stopband:
        btype = None
        if type(sb[-1]) is str:
            btype = sb[-1]
            sb = sb[:-1]

        # convert to normalize frequency
        sb = np.array(sb,dtype=np.float32)
        sb[sb<0] = (fs/2)+sb[sb<0]+1 # neg freq count back from nyquist
        Wn  = sb/(fs/2)

        if  Wn[1] < .0001 or .9999 < Wn[0]: # no filter
            continue

        # identify type from frequencies used, cliping if end of frequency range
        if Wn[0] < .0001:
            Wn = Wn[1]
            btype = 'highpass' if btype is None or btype == 'bandstop' else 'lowpass'
        elif .9999 < Wn[1]:
            Wn = Wn[0]
            btype = 'lowpass' if btype is None or btype == 'bandstop' else 'highpass'

        elif btype is None: # .001 < Wn[0] and Wn[1] < .999:
            btype = 'bandstop'

        if verb>0: print("{}={}={}".format(btype,sb,Wn))

        if ftype == 'butter':
            sosi = butter(order, Wn, btype=btype, output='sos')
        elif ftype == 'bessel':
            sosi = bessel(order, Wn, btype=btype, output='sos', norm='phase')
        else:
            raise ValueError("Unrecognised filter type")

        sos.append(sosi)

    # single big filter cascade
    sos = np.concatenate(sos,axis=0)
    return sos

def butter_sosfilt(X, stopband, fs:float, order:int=6, axis:int=-2, zi=None, verb=True, ftype='butter'):
    """use a (cascade of) butterworth SOS filter(s) filter X along axis

    Args:
        X (np.ndarray): the data to be filtered
        stopband ([type]): the filter band specifications in Hz, as a list of lists of stopbands (given as (low-pass,high-pass)) or pass bands (given as (low-cut,high-cut,'bandpass'))
        fs (float): the sampling rate of X
        order (int, optional): the desired filter order. Defaults to 6.
        axis (int, optional): the axis of X to filter along. Defaults to -2.
        zi ([type], optional): the internal filter state -- propogate between calls for incremental filtering. Defaults to None.
        verb (bool, optional): Verbosity level for logging. Defaults to True.
        ftype (str, optional): The type of filter to make, one-of: 'butter', 'bessel'. Defaults to 'butter'.

    Returns:
        X [np.ndarray]: the filtered version of X
        sos (np.ndarray): the designed filter coefficients
        zi (np.ndarray): the filter state for propogation between calls

    """    '''  '''
    if stopband is None: # deal with no filter case
        return (X,None,None)
    if axis < 0: # no neg axis
        axis = X.ndim+axis
    # TODO []: auto-order determination?
    sos = iir_sosfilt_sos(stopband, fs, order, passband=passband, ftype=ftype)
    sos = sos.astype(X.dtype) # keep as single precision

    if axis == X.ndim-2 and zi is None:
        zi = sosfilt_zi(sos) # (order,2)
        zi.astype(X.dtype)
        zi = sosfilt_zi_warmup(zi, X, axis, sos)

    else:
        zi = None
        print("Warning: not warming up...")

    # Apply the warmed up filter to the input data
    #print("zi={}".format(zi.shape))
    if not zi is None:
        #print("filt:zi X{} axis={}".format(X.shape,axis))
        X, zi  = sosfilt(sos, X, axis=axis, zi=zi)
    else:
        print("filt:no-zi")
        X  = sosfilt(sos, X, axis=axis) # zi=zi)

    # return filtered data, filter-coefficients, filter-state
    return (X, sos, zi)

def save_butter_sosfilt_coeff(filename=None, stopband=((45,65),(5.5,25,'bandpass')), fs=200, order=6, ftype='butter'):
    ''' design a butterworth sos filter cascade and save the coefficients '''
    import pickle
    sos = iir_sosfilt_sos(stopband, fs, order, passband=None, ftype=ftype)
    zi = sosfilt_zi(sos)
    if filename is None:
        # auto-generate descriptive filename
        filename = "{}_stopband{}_fs{}.pk".format(ftype,stopband,fs)
    print("Saving to: {}\n".format(filename))
    with open(filename,'wb') as f:
        pickle.dump(sos,f)
        pickle.dump(zi,f)
        f.close()

def test_butter_sosfilt():
    fs= 100
    X = np.random.randn(fs*10,2)
    X = np.cumsum(X,0)
    X = X + np.random.randn(1,X.shape[1])*100 # include start shift
    import matplotlib.pyplot as plt

    plt.clf();plt.subplot(511);plt.plot(X);

    pbs=(((0,1),(40,-1)),(10,-1),((5,10),(15,20),(45,50)))
    for i,pb in enumerate(pbs):
        Xf,_,_ = butter_sosfilt(X,pb,fs)
        plt.subplot(5,1,i+2);plt.plot(Xf);plt.title("{}".format(pb))

    # test incremental application
    pb=pbs[0]
    sos=None
    zi =None
    Xf=[]
    for i in range(0,X.shape[0],fs):
        if sos is None:
            # init filter and do 1st block
            Xfi,sos,zi = butter_sosfilt(X[i:i+fs,:],pb,fs,axis=-2)
        else: # incremenally apply
            Xfi,zi = sosfilt(sos,X[i:i+fs,:],axis=-2,zi=zi)
        Xf.append(Xfi)
    Xf = np.concatenate(Xf,axis=0)
    plt.subplot(5,1,5);plt.plot(Xf);plt.title("{} - incremental".format(pb))
    plt.show()

    # test diff specifications
    pb = ((0,1),(40,-1)) # pair stops
    Xf0,_,_ = butter_sosfilt(X,pb,fs,axis=-2)
    plt.subplot(3,1,1);plt.plot(Xf0);plt.title("{}".format(pb))
 
    pb = (1,40,'bandpass') # single pass
    Xfi,_,_ = butter_sosfilt(X,pb,fs,axis=-2)
    plt.subplot(3,1,2);plt.plot(Xfi);plt.title("{}".format(pb))

    pb = (1,40,'bandpass') # single pass
    Xfi,_,_ = butter_sosfilt(X,pb,fs,axis=-2,ftype='bessel')
    plt.subplot(3,1,3);plt.plot(Xfi);plt.title("{} - bessel".format(pb))
    plt.show()


# TODO[] : cythonize?
# TODO[X] : vectorize over d? ---- NO. 2.5x *slower*
def sosfilt_2d_py(sos,X,axis=-2,zi=None):
    ''' pure python fallback for second-order-sections filter in case scipy isn't available '''
    X = np.asarray(X)
    sos = np.asarray(sos)

    if zi is None:
        returnzi = False
        zi = np.zeros((sos.shape[0],2,X.shape[-1]),dtype=X.dtype)
    else:
        returnzi = True
        zi = np.asarray(zi)

    Xshape = X.shape
    if not X.ndim == 2:
        print("Warning: X>2d.... treating as 2d...")
        X = X.reshape((-1,Xshape[-1]))

    if axis < 0:
        axis = X.ndim + axis

    if not axis == X.ndim-2:
        raise ValueError("Only for time in dim 0/-2")

    if sos.ndim != 2 or sos.shape[1] != 6:
        raise ValueError('sos must be shape (n_sections, 6)')

    if zi.ndim != 3 or zi.shape[1] != 2 or zi.shape[2] != X.shape[1]:
        raise ValueError('zi must be shape (n_sections, 2, dim)')

    # pre-normalize sos if needed
    for j in range(sos.shape[0]):
        if sos[j,3] != 1.0:
            sos[j,:] = sos[j,:]/sos[j,3]
    
    n_signals = X.shape[1]
    n_samples = X.shape[0]
    n_sections = sos.shape[0]

    # extract the a/b
    b = sos[:,:3]
    a = sos[:,4:]

    # loop over outputs
    x_n = 0
    for i in range(n_signals):
        for n in range(n_samples):
            for s in range(n_sections):
                x_n = X[n, i]
                # use direct II transposed structure
                X[n, i] = b[s, 0] * x_n + zi[s, 0, i]
                zi[s, 0, i] = b[s, 1] * x_n - a[s, 0] * X[n, i] + zi[s, 1, i]
                zi[s, 1, i] = b[s, 2] * x_n - a[s, 1] * X[n, i]

    # back to input shape
    if not len(Xshape) == 2:
        X = X.reshape(Xshape)

    # match sosfilt, only return zi if given zi
    if returnzi :
        return X, zi
    else:
        return X

def sosfilt_zi_py(sos):
    ''' compute an initial state for a second-order section filter '''
    sos = np.asarray(sos)
    if sos.ndim != 2 or sos.shape[1] != 6:
        raise ValueError('sos must be shape (n_sections, 6)')

    n_sections = sos.shape[0]
    zi = np.empty((n_sections, 2))
    scale = 1.0
    for section in range(n_sections):
        b = sos[section, :3]
        a = sos[section, 3:]
        
        if a[0] != 1.0:
            # Normalize the coefficients so a[0] == 1.
            b = b / a[0]
            a = a / a[0]
        IminusA = np.eye(n_sections - 1) - np.linalg.companion(a).T
        B = b[1:] - a[1:] * b[0]
        # Solve zi = A*zi + B
        lfilter_zi = np.linalg.solve(IminusA, B)
        zi[section] = scale * lfilter_zi
        scale *= b.sum() / a.sum()
        
    return zi

def test_sosfilt_py():
    import pickle
    with open('butter_stopband((0, 5), (25, -1))_fs200.pk','rb') as f:
        sos = pickle.load(f)
        zi = pickle.load(f)

    X = np.random.randn(10000,3)
    print("X={} sos={}".format(X.shape,sos.shape))

    Xsci = sosfilt(sos,X.copy(),-2)
    Xpy = sosfilt_2d_py(sos,X.copy(),-2)

    import matplotlib.pyplot as plt
    plt.clf()
    plt.subplot(411);plt.plot(X[:500,:]);plt.title('X')
    plt.subplot(412);plt.plot(Xsci[:500,:]);plt.title('Xscipy')
    plt.subplot(413);plt.plot(Xpy[:500,:]);plt.title('Xpy')
    plt.subplot(414);plt.plot(Xsci-Xpy);plt.title('Xsci - Xpy')

# def butter_py(order,fc,fs,btype,output):
#     ''' pure python butterworth filter synthesis '''
#     if fc>=fs/2:
#         error('fc must be less than fs/2')

#     # I.  Find poles of analog filter
#     k= np.arange(order)
#     theta= (2*k -1)*np.pi/(2*order);
#     pa= -sin(theta) + j*cos(theta);     # poles of filter with cutoff = 1 rad/s
#     #
#     # II.  scale poles in frequency
#     Fc= fs/np.pi * tan(np.pi*fc/fs);          # continuous pre-warped frequency
#     pa= pa*2*np.pi*Fc;                     # scale poles by 2*pi*Fc
#     #
#     # III.  Find coeffs of digital filter
#     # poles and zeros in the z plane
#     p= (1 + pa/(2*fs))/(1 - pa/(2*fs))      # poles by bilinear transform
#     q= -np.ones((1,N));                   # zeros
#     #
#     # convert poles and zeros to polynomial coeffs
#     a= poly(p);                   # convert poles to polynomial coeffs a
#     a= real(a);
#     b= poly(q);                   # convert zeros to polynomial coeffs b
#     K= sum(a)/sum(b);             # amplitude scale factor
#     b= K*b;    

if __name__=='__main__':
    save_butter_sosfilt_coeff("sos_filter_coeff.pk")
    #test_butter_sosfilt()



import serial
from serial import Serial

from threading import Timer
import time
import logging
import sys
import struct
import numpy as np
import atexit
import datetime
import glob
# Define variables
SAMPLE_RATE = 250.0  # Hz
START_BYTE = 0xA0  # start of data packet
END_BYTE = 0xC0  # end of data packet

class OpenBCICyton(object):
    """ OpenBCICyton handles the connection to an OpenBCI Cyton board.

    The OpenBCICyton class interfaces with the Cyton Dongle and the Cyton board to parse the data received and output it to Python as a OpenBCISample object.

    Args:
        port: A string representing the COM port that the Cyton Dongle is connected to. e.g for Windows users 'COM3', for MacOS or Linux users '/dev/ttyUSB1'. If no port is specified it will try to find the first port available.

        daisy: A boolean indicating if there is a Daisy connected to the Cyton board.

        baud: An integer specifying the baudrate of the serial connection. The maximum baudrate of the Cyton board is 115200.

        timeout: An float specifying the maximum milliseconds to wait for serial data.

        max_packets_skipped: An integer specifying how many packets can be dropped before attempting to reconnect.

    """
    def __init__(self, port=None, daisy=False, baud=115200, timeout=None, max_packets_skipped=1):
        self._logger = logging.getLogger(self.__class__.__name__)

        self.baud = baud
        self.timeout = timeout
        self.daisy = daisy
        self.max_packets_skipped = max_packets_skipped
        self.streaming = False
        if port:
            self.port = port
        else:
            self.port = self.find_port()
        self.start_time = datetime.datetime.now().strftime("%Y-%m-%d_%H%M%S")
        if self.daisy:
            self.board_type = "CytonDaisy"
        else:
            self.board_type = "Cyton"

        # Connecting to the board
        self.ser = Serial(port=self.port, baudrate=self.baud, timeout=self.timeout)

        self._logger.info("Serial established.")

        # Perform a soft reset of the board
        time.sleep(2)
        self.ser.write(b'v')


        # wait for device to be ready
        time.sleep(1)

        if port != "loop://":
            self.print_incoming_text()


        self.packets_dropped = 0
        self.read_state = 0
        self.last_odd_sample = OpenBCISample(-1, [], [], self.start_time, self.board_type)  # used for daisy


        # Disconnects from board when terminated
        atexit.register(self.disconnect)

    def disconnect(self):
        """Disconnects the OpenBCI Serial."""
        if self.ser.isOpen():
            self._logger.info("Closing serial.")
            self.ser.close()

    def find_port(self):
        """Finds the port to which the Cyton Dongle is connected to."""
        # Find serial port names per OS
        if sys.platform.startswith('win'):
            ports = ['COM%s' % (i + 1) for i in range(256)]
        elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
            ports = glob.glob('/dev/ttyUSB*')
        elif sys.platform.startswith('darwin'):
            ports = glob.glob('/dev/tty.usbserial*')
        else:
            raise EnvironmentError('Error finding ports on your operating system')

        openbci_port = ''
        for port in ports:
            try:
                s = Serial(port=port, baudrate=self.baud, timeout=self.timeout)
                s.write(b'v')
                line = ''
                time.sleep(2)
                if s.inWaiting():
                    line = ''
                    c = ''
                    while '$$$' not in line:
                        c = s.read().decode('utf-8', errors='replace')
                        line += c
                    if 'OpenBCI' in line:
                        openbci_port = port
                s.close()
            except (OSError, serial.SerialException):
                pass
        if openbci_port == '':
            raise OSError('Cannot find OpenBCI port.')
        else:
            return openbci_port

    def stop_stream(self):
        """Stops Stream from the Cyton board."""
        self.streaming = False
        self.ser.write(b's')

    def reconnect(self):
        """Attempts to reconnect to the Cyton board if the connection was lost."""
        self.packets_dropped = 0
        self._logger.info("Reconnecting...")

        # Stop stream
        self.stop_stream()
        time.sleep(0.5)

        # Soft reset of the board
        self.ser.write(b'v')
        time.sleep(0.5)

        # Start stream
        self.ser.write(b'b')
        time.sleep(0.5)
        self.streaming = True

    def check_connection(self, max_packets_skipped=1, interval=2):
        """Verifies if the connection is stable. If not, it attempts to reconnect to the board"""
        if not self.streaming:
            self._logger.warning("Not streaming!")
            return

        # check number of dropped packets and reconnect if problem is too large
        elif self.packets_dropped > max_packets_skipped:
                #if error attempt to reconnect
                self.reconnect()

        # Check connection every 'interval' seconds
        Timer(interval, self.check_connection).start()


    def parse_board_data(self, maxbytes2skip=3000):
        """Parses the data from the Cyton board into an OpenBCISample object."""
        def read_board(n):
            bb = self.ser.read(n)
            if not bb:
                self._logger.warning("Device appears to be stalling. "
                                     "Quitting...")
                sys.exit()
                raise Exception('Device Stalled')
                sys.exit()
                return '\xFF'
            else:
                return bb

        for rep in range(maxbytes2skip):

            # Start Byte & ID
            if self.read_state == 0:
                b = read_board(1)

                if struct.unpack('B', b)[0] == START_BYTE:
                    if rep != 0:
                        self._logger.info(
                            "Skipped %d bytes before start found" % rep)
                        rep = 0

                    packet_id = struct.unpack('B', read_board(1))[0]
                    log_bytes_in = str(packet_id)

                    self.read_state = 1

            # Channel data
            elif self.read_state == 1:
                channels_data = []
                for c in range(8):
                    # Read 3 byte integers
                    literal_read = read_board(3)

                    unpacked = struct.unpack('3B', literal_read)
                    log_bytes_in = log_bytes_in + '|' + str(literal_read)

                    # Translate 3 byte int into 2s complement
                    if unpacked[0] > 127:
                        pre_fix = bytes(bytearray.fromhex('FF'))
                    else:
                        pre_fix = bytes(bytearray.fromhex('00'))

                    literal_read = pre_fix + literal_read

                    myInt = struct.unpack('>i', literal_read)[0]

                    # Append channel to channels data
                    channels_data.append(myInt)

                self.read_state = 2

            # Read Aux Data
            elif self.read_state == 2:
                aux_data = []
                for a in range(3):

                    acc = struct.unpack('>h', read_board(2))[0]
                    log_bytes_in = log_bytes_in + '|' + str(acc)

                    # Append to auxiliary data array
                    aux_data.append(acc)

                self.read_state = 3

            # Read End Byte
            elif self.read_state == 3:
                val = struct.unpack('B', read_board(1))[0]

                log_bytes_in = log_bytes_in + '|' + str(val)
                self.read_state = 0 # resets to read next packet

                if val == END_BYTE:
                    sample = OpenBCISample(packet_id, channels_data, aux_data, self.start_time, self.board_type)
                    self.packets_dropped = 0
                    return sample
                else:
                    self._logger.warning("ID:<%d> <Unexpected END_BYTE found <%s> instead of <%s>" % (packet_id, val, END_BYTE))
                    self.packets_dropped = self.packets_dropped + 1


    def write_command(self, command):
        """Sends string command to the Cyton board"""
        if command == '?':
            self.ser.write(command.encode())
            if self.ser.inWaiting():
                line = ''
                while '$$$' not in line:
                    line += self.ser.read().decode('utf-8', errors='replace')
                print(line)
        else:
            self.ser.write(command.encode())
            time.sleep(0.5)


    def start_stream(self, callback):
        """Start handling streaming data from the board. Call a provided callback for every single sample that is processed."""
        if not self.streaming:
            self.ser.write(b'b')
            self.streaming = True

        # Enclose callback function in a list
        if not isinstance(callback, list):
            callback = [callback]

        # checks connection
        self.check_connection(max_packets_skipped=self.max_packets_skipped)

        while self.streaming:

            #read current sample
            sample = self.parse_board_data()

            if not self.daisy:
                 for call in callback:
                     call(sample)

            # When daisy is connected wait to concatenate two samples
            else:
                # odd sample is daisy sample use later
                if ~sample.id % 2:
                    self.last_odd_sample = sample

                # Check if the next sample ID is concecutive, if not the packet is dropped
                elif sample.id - 1 == self.last_odd_sample.id:
                    # The auxiliary data is the average between the two samples.
                    avg_aux_data = list((np.array(sample.aux_data) + np.array(self.last_odd_sample.aux_data)) / 2)

                    sample_with_daisy = OpenBCISample(sample.id, sample.channels_data + self.last_odd_sample.channels_data, avg_aux_data, self.start_time, self.board_type)

                    for call in callback:
                        call(sample_with_daisy)
                        
    def print_incoming_text(self):
        """
        When starting the connection, print all the debug data until
        we get to a line with the end sequence '$$$'.
        """
        line = ''
        # Wait for device to send data
        time.sleep(1)

        if self.ser.inWaiting():
            line = ''
            c = ''
            # Look for end sequence $$$
            while '$$$' not in line:
                # we're supposed to get UTF8 text, but the board might behave otherwise
                c = self.ser.read().decode('utf-8',
                                           errors='replace')
                line += c
            self._logger.debug(line)
        else:
            self._logger.warning("No Message")


class OpenBCISample():
    """ Object that encapsulates a single sample from the OpenBCI board.

    Attributes:
        id: An int representing the packet id of the aquired sample.
        channels_data: An array with the data from the board channels.
        aux_data: An array with the aux data from the board.
        start_time: A string with the stream start time.
        board_type: A string specifying the board type, e.g 'cyton', 'daisy', 'ganglion'
    """

    def __init__(self, packet_id, channels_data, aux_data, init_time, board_type):
        self.id = packet_id
        self.channels_data = channels_data
        self.aux_data = aux_data
        self.start_time = init_time
        self.board_type = board_type



import os
import numpy as np

def getPosInfo(chnames=None,capFile='1010',overridechnms=0,prefixMatch=False,verb=0,capDir=None):
    """
     add electrode position info to a dimInfo structure
    
     cnames,pos2d,pos3d,iseeg=getPosInfo(di,capFile,overridechnms,prefixMatch,verb,capDir)
    
    Inputs:
      cnames -- channel names to get pos-info for
      capFile -- file name of a file which contains the pos-info for this cap
      overridechnms -- flag that we should ignore the channel names in di
      prefixMatch  -- [bool] match channel names if only the start matches?
      verb         -- [int] verbosity level  (0)
      capDir       -- 'str' directory to search for capFile
    Outputs:
      cnames -- [str] the names of the channels
      xy -- (nCh,2):float the 2d position of the channels
      xyz -- (nCh,2):float the 3d position of the channels
      iseeg - (nCh,):bool flag if this is an eeg channel, i.e. if it's name matched
    """
    cfnames,latlong,xy,xyz,cfile=readCapInf(capFile,capDir)
    
    if overridechnms  or (isinstance(cfnames,np.ndarray) and np.issubdtype(cfnames.dtype,np.number)) :
        if chnames is None:
            chnames = cfnames
        else:
            if len(chnames)<len(cfnames):
                chnames = cfnames[1:len(chnames)]
            else:
                chnames[1:len(cfnames)]=cfnames
                
    # Add the channel position info, and iseeg status
    # pre-allocate for the output
    pos2d=np.zeros((len(chnames),2),dtype=np.float)
    pos3d=np.zeros((len(chnames),3),dtype=np.float)
    iseeg=np.zeros((len(chnames)),  dtype=np.bool)
    cfmatched=np.zeros((len(cfnames),1),dtype=np.bool) # file channels matched
    for i,chnamei in enumerate(chnames):
        ti=None
        if isinstance(chnamei,str):
            for j,cfnamej in enumerate(cfnames):
                if verb>1 : print("Trying %s == %s?"%(chnamei,cfnamej),end='')
                if not cfmatched[j] and chnamei.lower()==cfnamej.lower():
                    if verb>1 : print("matched")
                    ti=j
                    cfmatched[j]=True
                    break
            if prefixMatch and ti is None:
                for j in range(0,numel(cfnames)):
                    if not cfmatched[j] and lower(cfnamej).startswith(lower(chnamei)):
                        ti=j
                        cfmatched[j]=True
                        break
        else:
            if (isinstance(chnamei,np.ndarray) and i < len(cfnames)):
                ti=chnamei
                cfmatched[chnamei]=True
            else:
                ti=None
                print('WARNING:: Channel names are difficult')

        # got the match, so update the info
        if (not ti is None and ti >= 0 and ti < len(cfnames)):
            if verb > 0:
                print('%3d) Matched : %s\t ->\t %s'%(i,chnames[i],cfnames[ti]))
            chnames[i] =cfnames[ti]
            pos2d[i,:]=xy[:,ti]
            pos3d[i,:]=xyz[:,ti]
            iseeg[i]  =True
            if any(np.isnan(pos2d[i,:])) or any(np.isnan(pos3d[i,:])) or any(np.isinf(pos2d[i,:])) or any(np.isinf(pos3d[i,:])):
                iseeg[i] = False
        else:
            pos2d[i,:] = (-1,1)
            pos3d[i,:] = (-1,1,0)
            iseeg[i] = False
    
    return (chnames,pos2d,pos3d,iseeg)

def testCase():
    import getPosInfo
    cn,xy,xyz,iseeg=getPosInfo.getPosInfo(['CPz','AFz'],'1010')

def readCapInf(cap='1010.txt',capRoots=None, verb=0):
    """     read a cap file
 
    Cname,latlong,xy,xyz,capfile=readCapInf(cap,capDir)
    
     Inputs:
    cap     -- file name of the cap-file
    capRoot -- directory(s) to look for caps in (['.',mfiledir,mfiledir'/positions','../../resources/caps/'])
    """

    if capRoots is None:
        filedir = os.path.dirname(os.path.abspath(__file__))
        capRoots = ('.',filedir,os.path.join(filedir,'positions'),os.path.join(filedir,'..','..','resources','caps'));    
    elif isinstance(capRoots,basestring):
        capRoots=(capRoots)

    capDir,capFnExt=os.path.split(cap)
    capFn,sep,capExt=capFnExt.partition('.') # get extn if given
    # search given directories for the capfile
    for cr in range(0,len(capRoots)):
        capRoot=capRoots[cr]
        if len(capExt)>0 : 
            capFile=os.path.join(capRoot,capDir,capFn+'.'+capExt)
            if verb > 0: print('Trying: ' + capFile)
            if os.path.exists(capFile):
                if verb > 0: print('success')
                break
        else:
            capFile=os.path.join(capRoot,capDir,capFn+'.txt')
            if verb > 0: print('Trying: ' + capFile)
            if os.path.exists(capFile):
                capExt='txt'
                if verb > 0: print('success')
                break
            capFile=os.path.join(capRoot,capDir,capFn+'.lay')
            if verb > 0: print('Trying: ' + capFile)
            if os.path.exists(capFile):
                capExt='lay'
                if verb > 0: print('success')
                break

    if not os.path.exists(capFile):
        raise ValueError('Couldnt find the capFile: '+cap)
    
    if 'xyz' in capExt:
        Cname,x,y,z= np.loadtxt(capFile,
                                dtype={'names': ('Cname', 'x', 'y', 'z'),
                                       'formats': ('|S12', 'float', 'float', 'float')},
                                unpack=True)
        #Cname,x,y,z=textread(capFile,'b'%s %f %f %f'',nargout=4)
        xyz=np.stack((x,y,z),1).T # [3 x nCh]
        xy=xyz2xy(xyz)
        latlong=xy2latlong(xy)
    else:
        if 'xy' in capExt:
            Cname,x,y,z= np.loadtxt(capFile,
                                    dtype={'names': ('Cname', 'x', 'y'),
                                           'formats': ('|S12', 'float', 'float')},
                                    unpack=True)
            #Cname,x,y=textread(capFile,'b'%s %f %f'',nargout=3)
            xy=np.stack((x,y),1).T #[2 x nCh]
            latlong=xy2latlong(xy)
            xyz=latlong2xyz(latlong)
        else:
            if 'lay' in capExt:
                tmp,x,y,w,h,Cname=np.loadtxt(capFile,
                                             dtype={'names': ('rownum', 'x', 'y', 'w','h','Cname'),
                                                    'formats': ('float', 'float', 'float', 'float', 'float', '|S12')},
                                             unpack=True)
                #ans,x,y,w,h,Cname=textread(capFile,'b'%d %f %f %f %f %s'',nargout=6)
                xy=np.stack((x,y),1).T # [2 x nCh]
                xy= xy - np.mean(xy,1)
                xy= xy / np.sqrt(np.mean(xy*xy,1))
                latlong=xy2latlong(xy)
                xyz=latlong2xyz(latlong)
            else:
                Cname,lat,lng=np.loadtxt(capFile,
                                          dtype={'names': ('Cname','lat','lng'),
                                                 'formats': ('|S12', 'float', 'float')},
                                             unpack=True)
                #Cname,lat,lng=textread(capFile,'b'%s %f %f'',nargout=3)
                latlong=np.stack((lat,lng),1).T # [2 x nCh]
                if np.max(np.abs(latlong[:])) > 2*np.pi:
                    latlong= latlong *np.pi / 180.0
                xyz=latlong2xyz(latlong)
                xy=latlong2xy(latlong)

    Cname = np.char.decode(Cname,'utf8')
                
    return Cname,latlong,xy,xyz,capFile
    
def xyz2xy(xyz):
    "convert 3d xyz coords to unrolled 2d xy positions"
    eegch=np.logical_not(np.any(np.logical_or(isnan(xyz),isinf(xyz)),0))
    cz=mean(xyz[3,eegch])
    r=np.abs(max(abs(xyz[3,eegch] - cz))*1.1)
    if r < eps:
        r=1
    h=xyz[2,eegch] - cz
    rr=np.sqrt(np.dot(2,(r*r - np.dot(r,h))) / (r*r-h*h))
    xy=np.zeros((2,xyz.shape[1]))
    xy[0,eegch]=xyz[0,eegch]*rr
    xy[1,eegch]=xyz[1,eegch]*rr
    xy[:,logical_not(eegch)]=np.NaN
    return xy
    
def xy2latlong(xy):
    "convert xy to lat-long, taking care to prevent division by 0"
    eegch=np.logical_not(np.any(np.logical_or(np.isnan(xy),np.isinf(xy)),0))
    latlong=np.zeros(np.shape(xy))
    latlong[0,eegch]= np.sqrt(np.sum(xy[:,eegch] ** 2,1))
    latlong[1,eegch]= np.atan2(xy[1,eegch],xy[0,eegch])
    latlong[:,np.logical_not(eegch)]=np.NaN
    return latlong
    
def latlong2xyz(latlong):
    "convert lat-long into 3-d x,y,z coords on the sphere"
    eegch=np.logical_not(np.any(np.logical_or(np.isnan(latlong),np.isinf(latlong)),0))
    xyz=np.zeros((3,latlong.shape[1]))
    xyz[0,eegch]=np.sin(latlong[0,eegch]) * np.cos(latlong[1,eegch])
    xyz[1,eegch]=np.sin(latlong[0,eegch]) * np.sin(latlong[1,eegch])
    xyz[2,eegch]=np.cos(latlong[0,eegch])
    xyz[:,np.logical_not(eegch)]=np.NaN
    return xyz
    
def latlong2xy(latlong):
    "convert lat-long to 2-d unrolled x,y coordinates"
    eegch=np.logical_not(np.any(np.logical_or(np.isnan(latlong),np.isinf(latlong)),0))
    xy=np.zeros((2,latlong.shape[1]))
    xy[0,eegch]=np.sin(latlong[0,eegch]) * np.cos(latlong[1,eegch])
    xy[1,eegch]=np.sin(latlong[0,eegch]) * np.sin(latlong[1,eegch])
    return xy
    



from mindaffectBCI.decoder.offline.read_mindaffectBCI import read_mindaffectBCI_message
from mindaffectBCI.utopiaclient import DataPacket
from time import sleep

class FileProxyHub:
    ''' Proxy UtopiaClient which gets messages from a saved log file '''
    def __init__(self, filename:str=None, speedup:float=None, use_server_ts:bool=True):
        self.filename = filename
        import glob
        import os
        if self.filename is None or self.filename == '-':
            # default to last log file if not given
            files = glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)),'../../logs/mindaffectBCI*.txt')) # * means all if need specific format then *.csv
            self.filename = max(files, key=os.path.getctime)
        else:
            files = glob.glob(os.path.expanduser(filename))
            self.filename = max(files, key=os.path.getctime)
        print("Loading : {}\n".format(self.filename))
        self.speedup = speedup
        self.isConnected = True
        self.lasttimestamp = None
        self.use_server_ts = use_server_ts
        self.file = open(self.filename,'r')
    
    def getTimeStamp(self):
        return self.lasttimestamp
    
    def autoconnect(self, *args,**kwargs):
        pass

    def sendMessage(self, msg):
        pass

    def getNewMessages(self, timeout_ms):
        msgs = []
        for line in self.file:
            msg = read_mindaffectBCI_message(line)
            if msg is None:
                continue
            # re-write timestamp to server time stamp
            if self.use_server_ts:
                msg.timestamp = msg.sts
            # initialize the time-stamp tracking
            if self.lasttimestamp is None or self.lasttimestamp==0: 
                self.lasttimestamp = msg.timestamp
            # add to outgoing message queue
            msgs.append(msg)
            # check if should stop
            if self.lasttimestamp is not None and self.lasttimestamp+timeout_ms < msg.sts:
                break
        else:
            # mark as disconneted at EOF
            self.isConnected = False
        if self.speedup :
            sleep(timeout_ms/1000./self.speedup)
        # update the time-stamp cursor
        self.lasttimestamp = self.lasttimestamp + timeout_ms
        return msgs

def testcase(filename, fs=200, fs_out=200, stopband=((45,65),(0,3),(25,-1)), order=4):
    import numpy as np
    from mindaffectBCI.decoder.UtopiaDataInterface import timestamp_interpolation, linear_trend_tracker, butterfilt_and_downsample

    U = FileProxyHub(filename)
    tsfilt = timestamp_interpolation(fs=fs,sample2timestamp=linear_trend_tracker(500))

    if stopband is not None:
        ppfn = butterfilt_and_downsample(stopband=stopband, order=order, fs=fs, fs_out=fs_out)
    else:
        ppfn = None
    
    #ppfn = None

    nsamp=0
    t=0
    data=[]
    ts=[]
    while U.isConnected:
        msgs = U.getNewMessages(100)
        print('.',end='',flush=True)
        for m in msgs:
            if m.msgID == DataPacket.msgID:
                timestamp = m.timestamp % (1<<24)
                samples = m.samples
                sample_ts = tsfilt.transform(timestamp,len(samples))
                if ppfn: # apply pre-processor
                    samples, sample_ts = ppfn.transform(samples, sample_ts[:,np.newaxis])
                    sample_ts = sample_ts[:,0]
                if len(samples) > 0:
                    data.extend(samples)
                    ts.extend(sample_ts)
    data = np.array(data)
    ts = np.array(ts)
    data = np.append(data,ts[:,np.newaxis],-1)
    # dump as pickle
    import pickle
    if ppfn is None:
        pickle.dump(dict(data=data),open('raw_fph.pk','wb'))
    else:
        pickle.dump(dict(data=data),open('pp_fph.pk','wb'))


if __name__=="__main__":
    import sys
    filename = sys.argv[1] if len(sys.argv)>1 else None

    filename = "C:\\Users\\Developer\\Downloads\\mark\\mindaffectBCI_brainflow_200911_1229_90cal.txt"

    testcase(filename)




import numpy as np
from mindaffectBCI.decoder.offline.read_mindaffectBCI import read_mindaffectBCI_messages
from mindaffectBCI.utopiaclient import StimulusEvent, DataPacket, ModeChange, NewTarget, Selection
import matplotlib.pyplot as plt
from mindaffectBCI.decoder.lower_bound_tracker import lower_bound_tracker

def timestampPlot(filename=None):
    import glob
    import os
    if filename is None or filename == '-':
        # default to last log file if not given
        files = glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)),'../../logs/mindaffectBCI*.txt')) # * means all if need specific format then *.csv
        filename = max(files, key=os.path.getctime)
    else:
        files = glob.glob(os.path.expanduser(filename))
        filename = max(files, key=os.path.getctime)
    print("Loading : {}\n".format(filename))    

    print('savefile={}'.format(filename))
    msgs = read_mindaffectBCI_messages(filename,regress=None) # load without time-stamp fixing.
    print("{} msgs".format(len(msgs)))
    dp = [ m for m in msgs if m.msgID == DataPacket.msgID]
    print("{} dp".format(len(dp)))
    samp = np.array([ m.samples.shape[0] for m in dp])
    svr = np.array([ m.sts for m in dp])
    rcv = np.array([ m.rts for m in dp])
    client = np.array([ m.timestamp for m in dp])
    samp = np.cumsum(samp)
    samp2ms = np.median(np.diff(svr)/np.diff(samp))

    lbt = lower_bound_tracker(a0=samp2ms)
    svr_filt = np.zeros(svr.shape)
    for i in range(samp.shape[0]):
        svr_filt[i] = lbt.transform(samp[i],svr[i])

    print("samp2ms {}".format(samp2ms))
    print("{}".format(samp.shape))
    print("Sample : {}".format(samp[:10]))
    print("Server : {}".format(svr[:10]))
    print("Reciev : {}".format(rcv[:10]))
    print("Client : {}".format(client[:10]))
    print("svr_flt: {}".format(svr_filt[:10]))
    print("{} samp = {}s".format(samp[-1],samp[-1]*samp2ms/1000))
    svr_err = svr - samp*samp2ms- svr[0]
    client_err = client - samp*samp2ms- client[0]
    svr_filt_err = svr_filt - samp*samp2ms- svr_filt[0]
    cent = np.median(svr_err); scale=np.median(np.abs(svr_err-cent))
    plt.plot(samp*samp2ms,svr_err,'.-',label='samp*samp2ms - server')
    plt.plot(samp*samp2ms,client_err,'.-',label='samp*samp2ms - client')
    plt.plot(samp*samp2ms,svr_filt_err,'.-',label='samp*samp2ms - filt(server)')
    plt.plot(samp*samp2ms,svr_err - client_err,'.-',label='server - client')
    plt.ylim((cent-scale*samp2ms,cent+scale*samp2ms))
    plt.xlabel('Time (ms)')
    plt.ylabel('time stamp error (ms)')
    plt.legend()
    plt.grid()
    plt.title('{}\nserver/client time-stamps w.r.t. fixed sample clock (@{}ms/samp)'.format(filename[-40:],samp2ms))
    plt.show()

if __name__=="__main__":
    #filename = "~/Desktop/trig_check/mindaffectBCI_*brainflow*.txt"
    #filename = '~/Desktop/mark/mindaffectBCI*ganglion*1411*.txt'
    #filename = '~/Desktop/pymindaffectBCI/logs/mindaffectBCI_*_201001_1859.txt'
    #filename = '~/Desktop/trig_check/mindaffectBCI_*brainflow2*.txt'
    filename = '~/Desktop/trig_check/mindaffectBCI_*timestamp*.txt'
    #filename=None
    #filename='~/Desktop/pymindaffectBCI/logs/mindaffectBCI_*_200928_2004.txt'; #mindaffectBCI_noisetag_bci_201002_1026.txt'
    timestampPlot(filename)




import os
import numpy as np
from scipy.io import loadmat
from scipy.signal import welch
from mindaffectBCI.decoder.utils import block_randomize, butter_sosfilt, upsample_codebook, lab2ind
from mindaffectBCI.decoder.multipleCCA import robust_whitener
from mindaffectBCI.decoder.updateSummaryStatistics import updateCxx

def load_mark_EMG(datadir, sessdir=None, sessfn=None, fs_out=60, stopband=((45,65),(0,10),(45,55),(95,105),(145,-1)), filterbank=None, verb=0, log=True, whiten=True, plot=False):

    fs=1000
    ch_names=None
    
    # load the data file
    Xfn = os.path.expanduser(datadir)
    if sessdir:
        Xfn = os.path.join(Xfn, sessdir)
    if sessfn:
        Xfn = os.path.join(Xfn, sessfn)
    sessdir = os.path.dirname(Xfn)

    print("Loading {}".format(Xfn))
    data = loadmat(Xfn)

    def squeeze(v):
        while v.size == 1 and v.ndim > 0:
            v = v[0]
        return v

    X = np.array([squeeze(d['buf']) for d in squeeze(data['data'])]) # ( nTrl, nch, nSamp)
    X = np.moveaxis(X,(0,1,2),(0,2,1)) # ( nTr, nSamp, nCh)
    X = np.ascontiguousarray(X) # ensure memory efficient
    lab = np.array([squeeze(e['value']) for e in data['devents']],dtype=int) # (nTrl,)

    import matplotlib.pyplot as plt
    if plot: plt.figure(100);plt.plot(X[0,:,:]);plt.title("raw")

    # preprocess -> spectral filter, in continuous time!
    if stopband is not None:
        if verb > 0:
            print("preFilter: {}Hz".format(stopband))
        X, _, _ = butter_sosfilt(X,stopband,fs)
        if plot:plt.figure(101);plt.plot(X[0,:,:]);plt.title("hp+notch+lp")
        # preprocess -> spatial whiten
        # TODO[] : make this  fit->transform method
        if whiten:
            print("spatial whitener")
            Cxx = updateCxx(None,X,None)
            W,_ = robust_whitener(Cxx)
            X = np.einsum("tsd,dw->tsw",X,W)
            if plot:plt.figure(102);plt.plot(X[0,:,:]);plt.title("+whiten")
            
    if not filterbank is None:
        if verb > 0:  print("Filterbank: {}".format(filterbank))
        # apply filter bank to frequency ranges into virtual channels
        Xs=[]
        # TODO: make a nicer shape, e.g. (tr,samp,band,ch)
        for bi,band in enumerate(filterbank):
            Xf, _, _ = butter_sosfilt(X,band,fs)
            Xs.append(Xf)
        # stack the bands as virtual channels
        X = np.concatenate(Xs,-1)
        
    X = np.abs(X) # rectify
    if log:
        print("log amplitude")
        X = np.log(np.maximum(X,1e-6))
    if plot:plt.figure(103);plt.plot(X[0,:,:]);plt.title("+abs")
    X, _, _ = butter_sosfilt(X,(40,-1),fs) # low-pass = envelope extraction
    if plot:plt.figure(104);plt.plot(X[0,:,:]);plt.title("env")
        
    # preprocess -> downsample @60hz
    resamprate=int(fs/fs_out)
    if resamprate > 1:
        if verb > 0:
            print("resample: {}->{}hz rsrate={}".format(fs,fs_out,resamprate))
        X = X[:, ::resamprate, :] # decimate X (trl, samp, d)
        fs = fs/resamprate

    # get Y
    Y_true, lab2class = lab2ind(lab) # (nTrl, e)
    Y_true = Y_true[:, np.newaxis, :] # ( nTrl,1,e)
    # TODO[] : exhaustive list of other targets...
    Yall = np.eye(Y_true.shape[-1],dtype=bool) # (nvirt,e)
    Yall = np.tile(Yall,(Y_true.shape[0],1,1)) # (nTrl,nvirt,e)
    Y = np.append(Y_true,Yall,axis=-2) # (nTrl,nvirt+1,e)
    # upsample to fs_out
    Y = np.tile(Y[:,np.newaxis,:,:],(1,X.shape[1],1,1)) #  (nTrl, nSamp, nY, e)
    Y = Y.astype(np.float32)
    
    # make coords array for the meta-info about the dimensions of X
    coords = [None]*X.ndim
    coords[0] = {'name':'trial'}
    coords[1] = {'name':'time','unit':'ms', \
                 'coords':np.arange(X.shape[1])*1000/fs, \
                 'fs':fs}
    coords[2] = {'name':'channel','coords':ch_names}
    # return data + metadata
    return (X, Y, coords)

def testcase():
    import sys

    sessfn = '~/data/bci/own_experiments/emg/facial/training_data_SV_6.mat'
    # command-line, for testing
    if len(sys.argv) > 1:
        sessfn = sys.argv[1]
 
    #from offline.load_mark_EMG import load_mark_EMG
    oX, oY, coords = load_mark_EMG(sessfn, fs_out=125, stopband=((0,10),(45,55),(95,105),(145,-1)), plot=False)
    fs = coords[1]['fs']
    ch_names = coords[2]['coords']
    X=oX.copy()
    Y=oY.copy()

    print("X({}){}".format([c['name'] for c in coords],X.shape))
    print("Y={}".format(Y.shape))
    print("fs={}".format(fs))

    # try as simple linear classification problem
    from sklearn.linear_model import Ridge, LogisticRegression, LogisticRegressionCV
    from sklearn.svm import LinearSVR, LinearSVC
    # trial average
    Xc = np.sum(X,axis=-2)
    Yc = np.argmax(Y[:,0,0,:],axis=-1)
    clsfr = LogisticRegression(C=1e8) # turn off regularization
    clsfr.fit(Xc,Yc)
    Yest = clsfr.predict(Xc)
    acc = sum([i==j for i, j in zip(Yc,Yest)]) / len(Yc); print("Acc:{}".format(acc))
    W = clsfr.coef_
    b = clsfr.intercept_
    Fe = np.einsum("Td,ed->Te",Xc,W) + b
    Fy = Fe
    Yest = np.argmax(Fy,-1)
    acc = sum([i==j for i, j in zip(Yc,Yest)]) / len(Yc); print("Acc:{}".format(acc))

    # per-sample
    clsfr = LogisticRegression(C=1e8)
    Xc = X.reshape((-1,X.shape[-1]))
    Yc = np.argmax(Y[...,0,:],axis=-1).reshape((-1,))
    clsfr.fit(Xc,Yc)
    W = clsfr.coef_
    b = clsfr.intercept_
    Fe = np.einsum("Td,ed->Te",Xc,W) + b
    Fy = Fe
    Yest = np.argmax(Fy,-1)
    acc = sum([i==j for i, j in zip(Yc,Yest)]) / len(Yc); print('Acc: {}'.format(acc))
    
    
    # visualize the dataset
    from analyse_datasets import debug_test_dataset
    debug_test_dataset(X, Y, coords, tau_ms=20, evtlabs=None, rank=15, model='cca')
    

    
    
if __name__=="__main__":
    testcase()



class ScreenStack(Screen):
    '''Special type of screen which contains a stack of screens to show,
    which are processed in stack order, i.e. LIFO'''
    def __init__(self,window):
        self.window=window
        self.screenStack=[]
    def draw(self,t):
        '''re-direct to the draw method of the top of the screen stack'''
        cur_screen=self.get()
        if cur_screen : 
            cur_screen.draw(t)
            # remove completed screens from the stack
            if cur_screen.is_done(): 
                self.pop()        
    def is_done(self):
        # only done when the screen stack is empty
        return not self.screenStack
    def push(self,screen):
        '''add to top of stack = run's first'''
        #print("push(%d)"%(len(self.screenStack))+str(screen))
        self.screenStack.append(screen)
    def pushback(self,screen):
        '''add to the bottom of the screen stack = runs's last'''
        #print("pushback(%d)"%(len(self.screenStack))+str(screen))
        self.screenStack.insert(0,screen)
    def pop(self):
        #print("pop(%d)"%(len(self.screenStack))+str(self.get()))
        return self.screenStack.pop()
    def get(self):
        return self.screenStack[-1] if self.screenStack else None



import numpy as np
from mindaffectBCI.decoder.utils import equals_subarray
def stim2event(M, evtypes=('re','fe'), axis=-1, oM=None):
    '''
    convert per-sample stimulus sequence into per-sample event sequence (e.g. rising/falling edge, or long/short flash)
    Inputs:
     M      - (...samp) or (...,samp,nY) for and/non-target features
     evnames - [nE]:str list of strings:
        "0", "1", "00", "11", "01" (aka. 're'), "10" (aka, fe), "010" (aka. short), "0110" (aka long)
        "nt"+evtname : non-target event, i.e. evtname occured for any other target
        "any"+evtname: any event, i.e. evtname occured for *any* target
        "rest" - not any of the other event types, N.B. must be *last* in event list
        "raw" - unchanged input intensity coding
        "grad" - 1st temporal derivative of the raw intensity
     axis - (-1) the axis of M which runs along 'time'
     oM - (...osamp) or (...,osamp,nY) prefix stimulus values of M, used to incrementally compute the  stimulus features
    Outputs:
     evt    - (M.shape,nE) event sequence
    Examples:
      #For a P300 bci, with target vs. non-target response use:
        M = np.array([[1,0,0,0,0,0,1,0],[0,0,1,0,0,0,0,0],[0,0,0,0,1,0,0,0]]).T
        E = stim2event(M,evtypes=['re','ntre'], axis=-2)
      #Or modeling as two responses, target-stim-response and any-stim-response
        E = stim2event(M,evtypes=('re','anyre'), axis=-2)
    '''
    if axis < 0: # ensure axis is positive!
        axis=M.ndim+axis
    if isinstance(evtypes, str):
        evtypes = [evtypes]
    if evtypes is None:
        return M[:,:,:,np.newaxis]
    if oM is not None:
        #  include the prefix
        M = np.append(oM,M,axis)

    # Copyright (c) MindAffect B.V. 2018
    E = np.zeros(M.shape+(len(evtypes), ), M.dtype) # list event types
    #print("E.dtype={}".format(E.dtype))
    if len(M) == 0: # guard empty inputs
        return E
    for ei, etype in enumerate(evtypes):
        modifier = None
        if etype.startswith("nt"):
            modifier = "nt"
            etype = etype[len(modifier):]
        if etype.startswith("any"):
            modifier = "any"
            etype = etype[len(modifier):]
            
        # 1-bit
        if etype == "flash" or etype == '1':
            F = (M == 1)
        elif etype == '0':
            F = (M == 0)
        # 2-bit
        elif etype == "00":
            F = equals_subarray(M, [0, 0], axis)
        elif etype == "re" or etype == "01":
            F = equals_subarray(M, [0, 1], axis)
        elif etype == "fe" or etype == "10":
            F = equals_subarray(M, [1, 0], axis)
        elif etype == "11":
            F = equals_subarray(M, [1, 1], axis)
        # 3-bit
        elif etype == "000":
            F = equals_subarray(M, [0, 0, 0], axis)
        elif etype == "001":
            F = equals_subarray(M, [0, 0, 1], axis)
        elif etype == "010" or etype == 'short':
            F = equals_subarray(M, [0, 1, 0], axis)
        elif etype == "011":
            F = equals_subarray(M, [0, 1, 1], axis)
        elif etype == "100":
            F = equals_subarray(M, [1, 0, 0], axis)
        elif etype == "101":
            F = equals_subarray(M, [1, 0, 1], axis)
        elif etype == "110":
            F = equals_subarray(M, [1, 1, 0], axis)
        elif etype == "111":
            F = equals_subarray(M, [1, 1, 1], axis)
        # 4-bit
        elif etype == "0110" or etype == 'long':
            F = equals_subarray(M, [0, 1, 1, 0], axis)
        # diff
        elif etype == "diff":
            F = np.diff(M, axis=axis) != 0
            # pad missing entry
            padshape=list(F.shape); padshape[axis] = 1
            F = np.append(np.zeros(padshape, dtype=F.dtype), F, axis)
        elif etype == "rest":
            F = np.logical_not(np.any(E, axis=-1))

        elif etype == 'raw':
            F = E
        elif etype == 'grad':
            F = np.diff(M,axis=axis)

        else:
            raise ValueError("Unrecognised evttype:{}".format(etype))

        # apply any modifiers wanted
        if modifier == "nt":
            # non-target, means true when OTHER targets are high, i.e. or over other outputs
            if not axis == M.ndim-2:
                raise ValueError("non-target only for axis==-2")
            anyevt = np.any(F > 0, axis=-1) # any stim event type
            for yi in range(F.shape[-1]):
                # non-target if target for *any* Y except this one
                F[..., yi] = np.logical_and(F[..., yi] == 0, anyevt)
                
        elif modifier == "any":
            if not axis == M.ndim-2:
                raise ValueError("any feature only for axis==-2")   
            # any, means true if any target is true, N.B. use logical_or to broadcast
            F = np.logical_or(F, np.any(F > 0, axis=-1, keepdims=True))

        E[..., ei] = F

    if oM is not None:
        # strip the prefix
        # build index expression to get the  post-fix along axis
        idx=[slice(None)]*E.ndim
        idx[axis]=slice(oM.shape[axis],E.shape[axis])
        # get  the postfix
        E = E[tuple(idx)]
    #print("E.dtype={}".format(E.dtype))
    return E

def testcase():
    from stim2event import stim2event
    # M = (samp,Y)
    M = np.array([[0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0],
                  [0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0]])
    
    print("Raw  :{}".format(M))
    e = stim2event(M, 'flash', axis=-1);     print("flash:{}".format(e[0, ...].T))
    e = stim2event(M, 're', axis=-1);        print("re   :{}".format(e[0, ...].T))
    e = stim2event(M, 'fe', axis=-1);        print("fe   :{}".format(e[0, ...].T))
    e = stim2event(M, 'diff', axis=-1);      print("diff :{}".format(e[0, ...].T))
    e = stim2event(M, 'ntre', axis=-1);      print("ntre :{}".format(e[0, ...].T))
    e = stim2event(M, ('re', 'fe'), axis=-1); print("refe :{}".format(e[0, ...].T))
    e = stim2event(M, ('re', 'fe', 'rest'), axis=-1); print("referest :{}".format(e[0, ...].T))

    # test incremental calling, propogating prefix between calls
    oM= None
    e = []
    for bi,b in enumerate(range(0,M.shape[-1],2)):
        bM = M[:,b:b+2]
        eb = stim2event(bM,('re','fe'),axis=-1,oM=oM)
        e.append(eb)
        oM=bM
    e = np.concatenate(e,-2)
    print("increfe :{}".format(e[0, ...].T))

if __name__ == "__main__":
    testcase()



import argparse
from pylsl import StreamInlet, resolve_stream, StreamInfo
from mindaffectBCI import utopiaclient 
from time import time, sleep
import traceback

PACKETRATE_HZ = 50
LOGINTERVAL_S = 3
t0=None
nextLogTime=None
def printLog(nSamp, nBlock):
    ''' textual logging of the data arrivals etc.'''
    global t0, nextLogTime
    t = time()
    if t0 is None: 
        t0 = t
        nextLogTime = t
    if t > nextLogTime:
        elapsed = time()-t0
        print("%d %d %f %f (samp,blk,s,hz)"%(nSamp, nBlock, elapsed, nSamp/elapsed), flush=True)
        nextLogTime = t +LOGINTERVAL_S


def parse_args():
    parser = argparse.ArgumentParser ()
    # use docs to check which parameters are required for specific board, e.g. for Cyton - set serial port
    parser.add_argument ('--host', type = str, help  = 'host name for the utopia hub', required = False, default = None)
    parser.add_argument ('--type', type = str, help  = 'stream type of the LSL resolver', required = False, default = 'EEG')
    args = parser.parse_args ()
    return args

inlet = None
client = None
def run (host=None, streamtype=None, **kwargs):
    global inlet, client

    # first resolve an EEG stream on the lab network
    print("looking for an EEG stream...")
    streams = resolve_stream('type', 'EEG')

    # create a new inlet to read from the stream
    inlet = StreamInlet(streams[0])

    # get meta-info about this stream
    info = inlet.info()
    print("stream info: {}".format(info.as_xml()))
    nch = info.channel_count()
    labels=[]
    try:
        ch = info.desc().child("channels").child("channel")
        for k in range(nch):
            labels.append(ch.child_value("label"))
            ch = ch.next_sibling()
    except:
        pass
    fSample = info.nominal_srate()
    print("board with {} ch @ {} hz".format(nch, fSample))

    # connect to the utopia client
    client = utopiaclient.UtopiaClient()
    client.disableHeartbeats() # disable heartbeats as we're a datapacket source
    client.autoconnect(host)
    # don't subscribe to anything
    client.sendMessage(utopiaclient.Subscribe(None, ""))
    print("Putting header.")
    client.sendMessage(utopiaclient.DataHeader(None, len(eeg_channels), fSample, ""))

    maxpacketsamples = int(32000 / 4 / len(eeg_channels))
    nSamp=0
    nBlock=0
    data=None
    ots = None
    while True:

        # grab all the new data + time-stamp
        samples, timestamps = inlet.pull_chunk(timeout=1/PACKETRATE_HZ)
        if len(samples) == 0:
            sleep(.001)
            continue

        # forward the *EEG* data to the utopia client
        nSamp = nSamp +len(samples)

        # fit time-stamp into 32-bit int (with potential wrap-around)
        ts = timestamps[-1]
        ts = (int(ts*1000))%(1<<31) 
        #ts = client.getTimeStamp()
        client.sendMessage(utopiaclient.DataPacket(ts, samples))
        nBlock = nBlock + 1

        # limit the packet sending rate..
        printLog(nSamp,nBlock)        

if __name__ == "__main__":
    #run(board_id=1,serial_port='com3') # ganglion
    run()

    args=parse_args()    
    try:
        run(**vars(args))
    except:
        traceback.print_exc()



#!/usr/bin/env python3
#  Copyright (c) 2019 MindAffect B.V. 
#  Author: Jason Farquhar <jason@mindaffect.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from mindaffectBCI.noisetag import Noisetag, sumstats
from psychopy import visual, core

# dictionary mapping from stimulus-state to colors
state2color={0:(-.8,-.8,-.8), # off=grey
             1:(1,1,1),    # on=white
             2:(-1,1,-1),    # cue=green
             3:(-1,-1,1)}    # feedback=blue
def draw(time):
    '''draw the display with colors from noisetag'''
    global lastfliptime, window
    # send info on the *previous* stimulus state, with the recorded vsync time (if available)
    fliptime = lastfliptime if lastfliptime is not None else nt.getTimeStamp()
    nt.sendStimulusState(timestamp=fliptime)
    # update and get the new stimulus state to display
    try : 
        nt.updateStimulusState()
        stimulus_state,target_state,objIDs,sendEvents=nt.getStimulusState()
    except StopIteration :
        exit() # terminate app when noisetag is done
        return

    # draw the display with the instructed colors
    if stimulus_state : 
        for i,s in enumerate(squares):
            s.fillColor = state2color[stimulus_state[i]]

    # some textual logging of what's happening
    if target_state is not None and target_state>=0:
        opto.fillColor = state2color[1 if target_state==1 else 0]
        print("*" if target_state>0 else '.',end='',flush=True)
    else:
        print('.',end='',flush=True)

# WARNING: Hacking ahead!!!
# record the flip time
lastfliptime=0

# define a trival selection handler
def selectionHandler(objID):
    print("Selected: %d"%(objID))    


# Initialize the noise-tagging connection
nt = Noisetag()
nt.connect(timeout_ms=5000)
nt.addSelectionHandler(selectionHandler)
# tell the noisetag framework to run a full : calibrate->prediction sequence
nt.setnumActiveObjIDs(2)
nt.startExpt(nCal=5,nPred=10,duration=4,framesperbit=1,cuedprediction=True)

# Initialize the drawing window
# make a default window, with fixed size for simplicty, and vsync for timing
if False:
    window = visual.Window(size=(640,480),color=(-1,-1,-1))
else:
    window = visual.Window(fullscr=True,color=(-1,-1,-1))

# grid
squares = [visual.Rect(window,pos=(-.5,0),size=(.4,.4),autoDraw=True),
           visual.Rect(window,pos=(.5,0),size=(.4,.4),autoDraw=True)]
# opto-sensor
opto = visual.Rect(window,pos=(-1,1),size=(.4,.4),autoDraw=True)

# run the main loop
clock = core.Clock()
while True:
    draw(clock.getTime())
    ffliptime = window.flip() # flip-time in seconds...
    lastfliptime = nt.getTimeStamp()
    #lastfliptime = ffliptime/1000.0 # don't use, need to fix hearbeats also!


import os
import numpy as np
from mindaffectBCI.decoder.offline.read_mindaffectBCI import read_mindaffectBCI_data_messages
from mindaffectBCI.decoder.devent2stimsequence import devent2stimSequence, upsample_stimseq
from mindaffectBCI.decoder.utils import block_randomize, butter_sosfilt, upsample_codebook, lab2ind, window_axis, unwrap
from mindaffectBCI.decoder.UtopiaDataInterface import butterfilt_and_downsample

def load_mindaffectBCI(datadir, sessdir=None, sessfn=None, fs_out=100, stopband=((45,65),(5.5,25,'bandpass')), order=6, ftype='butter', verb=0, iti_ms=1000, trlen_ms=None, offset_ms=(-500,500), **kwargs):
    
    # load the data file
    Xfn = datadir
    if sessdir:
        Xfn = os.path.join(Xfn, sessdir)
    if sessfn:
        Xfn = os.path.join(Xfn, sessfn)
    sessdir = os.path.dirname(Xfn)

    if verb >= 0: print("Loading {}".format(Xfn))
    # TODO []: convert to use the on-line time-stamp code
    X, messages = read_mindaffectBCI_data_messages(Xfn, **kwargs)

    # TODO[]: get the header if there is one?

    import pickle
    pickle.dump(dict(data=X),open('raw_lmbci.pk','wb'))

    # strip the data time-stamp channel
    data_ts = X[...,-1].astype(np.float64) # (nsamp,)
    data_ts = unwrap(data_ts)
    X = X[...,:-1] # (nsamp,nch)
    
    # estimate the sample rate from the data -- robustly?
    idx = range(0,data_ts.shape[0],1000)
    samp2ms = np.median( np.diff(data_ts[idx])/1000.0 ) 
    fs = 1000.0 / samp2ms
    ch_names = None

    if verb >= 0: print("X={} @{}Hz".format(X.shape,fs),flush=True)

    # pre-process: spectral filter + downsample
    # incremental call in bits
    ppfn = butterfilt_and_downsample(stopband=stopband, order=order, fs=fs, fs_out=fs_out, ftype=ftype)
    #ppfn = None
    if ppfn is not None:
        if verb >= 0:
            print("preFilter: {}th {} {}Hz & downsample {}->{}Hz".format(order,ftype,stopband,fs,fs_out))
        #ppfn.fit(X[0:1,:])
        # process in blocks to be like the on-line, use time-stamp as Y to get revised ts
        if False:
            idxs = np.arange(0,X.shape[0],6); idxs[-1]=X.shape[0]
            Xpp=[]
            tspp=[]
            for i in range(len(idxs)-1):
                idx = range(idxs[i],idxs[i+1])
                Xi, tsi = ppfn.transform(X[idx,:], data_ts[idx,np.newaxis])
                Xpp.append(Xi)
                tspp.append(tsi)
            X=np.concatenate(Xpp,axis=0)
            data_ts = np.concatenate(tspp,axis=0)
        else:
            X, data_ts = ppfn.transform(X, data_ts[:,np.newaxis])
        data_ts = data_ts[...,0] # map back to 1d
        fs  = ppfn.out_fs_ # update with actual re-sampled data rate
        #dur_s = (data_ts[-1] - data_ts[0])/1000.0
        #fs  = X.shape[0] / dur_s


    # extract the stimulus sequence
    Me, stim_ts, objIDs, _ = devent2stimSequence(messages)
    stim_ts = unwrap(stim_ts.astype(np.float64))

    import pickle
    pickle.dump(dict(data=np.append(X,data_ts[:,np.newaxis],-1),stim=np.append(Me,stim_ts[:,np.newaxis],-1)),open('pp_lmbci.pk','wb'))

    # up-sample to stim rate
    Y, stim_samp = upsample_stimseq(data_ts, Me, stim_ts, objIDs)
    Y_ts = np.zeros((Y.shape[0],),dtype=int); 
    Y_ts[stim_samp]=stim_ts
    if verb > 0: print("Y={} @{}Hz".format(Y.shape,fs),flush=True)

    # slice into trials
    # isi = interval *before* every stimulus --
    #  include data-start so includes 1st stimulus
    isi = np.diff(np.concatenate((data_ts[0:1],stim_ts,data_ts[-2:-1]),axis=0))
    #print('isi={}'.format(isi))
    # get trial indices in stimulus messages as sufficiently large inter-stimulus gap
    # N.B. this is the index in stim_ts of the *start* of the new trial
    trl_stim_idx = np.flatnonzero(isi > iti_ms)
    # get duration of stimulus in each trial
    trl_dur = stim_ts[trl_stim_idx[1:]-1] - stim_ts[trl_stim_idx[:-1]]
    print('trl_dur: {}'.format(trl_dur))
    # estimate the best trial-length to use
    if trlen_ms is None:
        trlen_ms = np.median(trl_dur)
    # strip any trial too much shorter than trlen_ms (50%)
    keep = np.flatnonzero(trl_dur>trlen_ms*.5)
    # re-compute the trlen_ms for the good trials
    trl_stim_idx = trl_stim_idx[keep]
    
    # compute the trial start/end relative to the trial-start
    trlen_samp  = int(trlen_ms *  fs / 1000)
    offset_samp = [int(o*fs/1000) for o in offset_ms]
    bgnend_samp = (offset_samp[0], trlen_samp+offset_samp[1]) # start end slice window
    xlen_samp = bgnend_samp[1]-bgnend_samp[0]
    
    # get the trial starts as indices & ms into the data array
    trl_samp_idx = stim_samp[trl_stim_idx]
    trl_ts       = stim_ts[trl_stim_idx]
    
    # extract the slices
    Xraw = X.copy()
    Yraw = Y.copy()
    X = np.zeros((len(trl_samp_idx), xlen_samp, Xraw.shape[-1]), dtype=Xraw.dtype) # (nTrl,nsamp,d)
    Y = np.zeros((len(trl_samp_idx), xlen_samp, Yraw.shape[-1]), dtype=Yraw.dtype)
    Xe_ts = np.zeros((len(trl_samp_idx), xlen_samp),dtype=int)
    Ye_ts = np.zeros((len(trl_samp_idx), xlen_samp),dtype=int)
    ep_idx = np.zeros((len(trl_samp_idx), xlen_samp),dtype=int)
    print("slicing {} trials =[{} - {}] samples @ {}Hz".format(len(trl_samp_idx),bgnend_samp[0], bgnend_samp[1],fs))
    for ti, si in enumerate(trl_samp_idx):
        idx = range(si+bgnend_samp[0],min(Xraw.shape[0],si+bgnend_samp[1]))
        nsamp = len(idx) #min(si+bgnend_samp[1],Xraw.shape[0])-(si+bgnend_samp[0])
        X[ti, :nsamp, :] = Xraw[idx, :]
        Xe_ts[ti,:nsamp] = data_ts[idx]
        Y[ti, :nsamp, :] = Yraw[idx, :]
        Ye_ts[ti,:nsamp] = Y_ts[idx]
        ep_idx[ti,:nsamp] = list(idx)
        
    del Xraw, Yraw
    if verb > 0: print("X={}\nY={}".format(X.shape,Y.shape))

    import pickle
    pickle.dump(dict(X=X,Y=Y,X_ts=Xe_ts,Y_ts=Ye_ts,ep_idx=ep_idx),open('sliced_lmbci.pk','wb'))


    # make coords array for the meta-info about the dimensions of X
    coords = [None]*X.ndim
    coords[0] = dict(name='trial', coords=trl_ts, trl_idx=ep_idx, trl_ts=Xe_ts, Y_ts=Ye_ts)
    coords[1] = {'name':'time','unit':'ms', \
                 'coords':np.arange(X.shape[1])/fs, \
                 'fs':fs}
    coords[2] = dict(name='channel', coords=ch_names)
    # return data + metadata
    return (X, Y, coords)

def testcase():
    import sys

    if len(sys.argv)>1:
        sessfn = sys.argv[1]
    else:
        # default to last log file if not given
        import glob
        import os
        fileregexp = '../../../logs/mindaffectBCI*.txt'
        #fileregexp = '../../../../utopia/java/utopia2ft/UtopiaMessages*.log'
        files = glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)),fileregexp)) # * means all if need specific format then *.csv
        sessfn = max(files, key=os.path.getctime)
 
    #sessfn = "C:\\Users\\Developer\\Downloads\\mark\\mindaffectBCI_brainflow_200911_1229_90cal.txt"
    from mindaffectBCI.decoder.offline.load_mindaffectBCI import load_mindaffectBCI
    print("Loading: {}".format(sessfn))
    X, Y, coords = load_mindaffectBCI(sessfn, fs_out=100, regress=False)
    times = coords[1]['coords']
    fs = coords[1]['fs']
    ch_names = coords[2]['coords']

    print("X({}){}".format([c['name'] for c in coords],X.shape))
    print("Y={}".format(Y.shape))
    print("fs={}".format(fs))

    # visualize the dataset
    from mindaffectBCI.decoder.analyse_datasets import debug_test_dataset
    debug_test_dataset(X[:,:,:], Y[:,:,:], coords,
                        preprocess_args=dict(badChannelThresh=None, badTrialThresh=3, stopband=None, whiten_spectrum=False, whiten=False),
                        tau_ms=450, evtlabs=('re','fe'), rank=1, model='cca')
    
    
if __name__=="__main__":
    testcase()
